#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os


def ignore_exc(func):
    # noinspection PyBroadException
    try:
        func()
    except:
        pass


def crc16(data, init):
    crc = init
    for bb in data:
        for _ in range(8):
            if (bb ^ crc) & 1:
                crc = (crc >> 1) ^ 0x8408
            else:
                crc >>= 1
            bb >>= 1
    return crc


def rotate_bit16_left(val, nrolls):
    while nrolls:
        bit = (val & (1 << 15)) >> 15
        val <<= 1
        val |= bit
        nrolls -= 1
    return val & 0xffff


def checksum(data, init):
    assert ((len(data) % 2) == 0)
    cksum = init
    for ii in range(0, len(data), 2):
        cksum += ((data[ii] << 8) | data[ii + 1])
        cksum = rotate_bit16_left(cksum & 0xffff, 2)
    return cksum


def byte_to_char(val):
    if 0x20 <= val <= 0x7d:
        return chr(val)
    return ' '


def bits_from_int16(data, index, start_bit, end_bit):
    val = 0
    int16 = (data[index] << 8) | data[index + 1]
    for nbit in range(start_bit, end_bit + 1):
        if int16 & (1 << nbit):
            val += (1 << (nbit - start_bit))
    return val


def int16_to_bits(data, index, start_bit, end_bit, value):
    nbits = (end_bit - start_bit + 1)
    if value < 0:
        raise Exception('value must be >= 0')
    if value >= (2 ** nbits):
        raise Exception('value (%d) too large for bits (%d)' % (value, nbits))
    value <<= start_bit
    int16 = (data[index] << 8) | data[index + 1]
    for nbit in range(start_bit, end_bit + 1):
        if value & (1 << nbit):
            int16 |= (1 << nbit)
        else:
            int16 &= ~(1 << nbit)
    data[index] = (int16 >> 8)
    data[index + 1] = (int16 & 0xff)


def str_from_props(instance):
    strs = ['%s: %s' % (pp, getattr(type(instance), pp).__get__(instance, type(instance)))
            for pp in dir(type(instance)) if
            not pp.startswith('_') and isinstance(getattr(type(instance), pp), property)]
    return os.linesep.join(sorted(strs))


def str_from_data(data, start, end):
    chars = []
    for ii in range(start, end + 2, 2):
        chars.append(byte_to_char(data[ii + 1]))
        chars.append(byte_to_char(data[ii]))
    return ''.join(chars)


def swap_bytes(data, start=0):
    for ii in range(start, len(data), 2):
        tmp = data[ii]
        data[ii] = data[ii + 1]
        data[ii + 1] = tmp
    return data


def set_bit(int16, bit, val):
    if val:
        int16 |= (1 << bit)
    else:
        int16 &= ~(1 << bit)
    return int16



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from bumpemu.controller import constants
from bumpemu.controller.serialize import append_uint16, append_array, append_bool


class ChargerSettings(object):
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self.port_number = 0
        self.requested_operation = constants.ChargerOperation.NONE
        self.requested_chemistry = constants.Chemistry.NONE
        self.requested_cell_count = 0
        self.requested_ir = 0
        self.requested_capacity = 0
        self.requested_charge_c = 0
        self.requested_discharge_c = 0
        self.requested_charge_rate = 0
        self.requested_discharge_rate = 0
        self.requested_charge_cutoff_cell_volts = 0
        self.requested_discharge_cutoff_cell_volts = 0
        self.requested_fuel_curve = [0] * 11
        self.multi_charger_mode = 0
        self.power_supply_mode = constants.PowerSupplyMode.DC
        self.use_balance_leads = True

    def serialize(self):
        buf = bytearray()
        buf.append(self.port_number)
        # noinspection PyTypeChecker
        buf.append(self.requested_operation.value)
        # noinspection PyTypeChecker
        buf.append(self.requested_chemistry.value)
        buf.append(self.requested_cell_count)
        append_uint16(buf, int(round(self.requested_ir * 100)))
        append_uint16(buf, self.requested_capacity)
        append_uint16(buf, int(round(self.requested_charge_c * 10)))
        append_uint16(buf, int(round(self.requested_discharge_c * 10)))
        append_uint16(buf, self.requested_charge_rate)
        append_uint16(buf, self.requested_discharge_rate)
        append_uint16(buf, int(round(self.requested_charge_cutoff_cell_volts * 1000)))
        append_uint16(buf, int(round(self.requested_discharge_cutoff_cell_volts * 1000)))
        fuel_curve = [int(round(val * 1000)) for val in self.requested_fuel_curve]
        append_array(buf, fuel_curve, append_uint16)
        buf.append(self.multi_charger_mode)
        # noinspection PyTypeChecker
        buf.append(self.power_supply_mode.value)
        append_bool(buf, self.use_balance_leads)
        return buf



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from bumpemu.controller.serialize import append_uint16, append_array, append_bool, append_str


class BumpSettings(object):
    MAX_NAME_LEN = 16

    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self.volume_level = 0
        self.touch_cal_dx = 0
        self.touch_cal_dy = 0
        self.touch_cal_cx = 0
        self.touch_cal_cy = 0
        self.custom_color_idle = 0
        self.custom_color_active = 0
        self.custom_color_complete = 0
        self.custom_color_safety = 0
        self.custom_color_setup = 0
        self.selected_color_theme = 0
        self.screen_layout = 0
        self.last_bluetooth_uuid = [0, 0, 0, 0, 0, 0]
        self.cell_ir_warning_threshold = 0
        self.capacity_warning_threshold = 0
        self.presets_enabled = False
        self.cycle_graph_caching_enabled = False
        self._charger_ports_disabled = [True, True, True, True]
        self.touch_calibration_redone = False
        self._power_sources = [0, 0, 0, 0]
        self._wiring_modes = [0, 0, 0, 0]
        self.charger_upgrade_states = [0, 0, 0, 0]
        self.charger_upgrade_models = [0, 0, 0, 0]
        self.power_source_defaults_created = False
        self._power_source_names = [None, None, None, None]
        self._power_source_types = [0, 0, 0, 0]
        self._power_source_warn_dod = [True, True, True, True]
        self._power_source_low_volts = [0, 0, 0, 0]
        self._power_source_max_amps = [0, 0, 0, 0]
        self._power_source_max_regen_amps = [0, 0, 0, 0]
        self._power_source_max_regen_volts = [0, 0, 0, 0]
        self._power_source_regen_dchg_enabled = [True, True, True, True]
        self.power_source_initial_setup_complete = True
        self.device_name = 'foobar'
        self.checksum = 0

    def set_power_source_params(self, index, name, typ, low_volts, max_amps,
                                warn_dod=True, max_regen_amps=0, max_regen_volts=0, regen_dchg_enabled=False):
        self._power_source_names[index] = name
        self._power_source_types[index] = typ
        self._power_source_warn_dod[index] = warn_dod
        self._power_source_low_volts[index] = low_volts
        self._power_source_max_amps[index] = max_amps
        self._power_source_max_regen_amps[index] = max_regen_amps
        self._power_source_max_regen_volts[index] = max_regen_volts
        self._power_source_regen_dchg_enabled[index] = regen_dchg_enabled

    def set_power_source(self, port, index):
        self._power_sources[port] = index

    def set_xh_wiring_mode(self, port):
        self._wiring_modes[port] = 1

    def enable_charger_port(self, port):
        self._charger_ports_disabled[port] = False

    def serialize(self):
        buf = bytearray()
        buf.append(self.volume_level)
        append_uint16(buf, self.touch_cal_dx)
        append_uint16(buf, self.touch_cal_dy)
        append_uint16(buf, self.touch_cal_cx)
        append_uint16(buf, self.touch_cal_cy)
        append_uint16(buf, self.custom_color_idle)
        append_uint16(buf, self.custom_color_active)
        append_uint16(buf, self.custom_color_complete)
        append_uint16(buf, self.custom_color_safety)
        append_uint16(buf, self.custom_color_setup)
        buf.append(self.selected_color_theme)
        buf.append(self.screen_layout)
        append_array(buf, self.last_bluetooth_uuid)
        buf.append(self.cell_ir_warning_threshold)
        buf.append(self.capacity_warning_threshold)
        append_bool(buf, self.presets_enabled)
        append_bool(buf, self.cycle_graph_caching_enabled)
        append_array(buf, self._charger_ports_disabled, append_bool)
        buf.append(0)
        buf.append(self.touch_calibration_redone)
        buf.append(0)
        append_array(buf, self._power_sources)
        append_array(buf, [0] * 4)
        append_array(buf, self._wiring_modes)
        append_array(buf, self.charger_upgrade_states)
        append_array(buf, self.charger_upgrade_models)
        append_bool(buf, self.power_source_defaults_created)
        append_array(buf, self._power_source_names, lambda buff, val: append_str(buff, val, self.MAX_NAME_LEN))
        append_array(buf, self._power_source_types)
        append_array(buf, self._power_source_warn_dod, append_bool)
        append_array(buf, self._power_source_low_volts, append_uint16)
        append_array(buf, self._power_source_max_amps, append_uint16)
        append_array(buf, self._power_source_max_regen_amps, append_uint16)
        append_array(buf, self._power_source_max_regen_volts, append_uint16)
        append_array(buf, self._power_source_regen_dchg_enabled, append_bool)
        append_bool(buf, self.power_source_initial_setup_complete)
        append_str(buf, self.device_name, self.MAX_NAME_LEN)
        append_array(buf, [0] * 70)
        append_uint16(buf, self.checksum)
        return buf



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import logging
import argparse
from yaml import load, Loader
import dbus
import dbus.mainloop.glib
import serial.tools.list_ports as list_ports
from gi.repository import GLib
from bumpemu.controller import bluez_dbus, constants
from bumpemu.controller.emulator import BumpEmulator, UartAdvertisement
from bumpemu.controller.messages.battery import Battery
from bumpemu.charger.powerlab import Powerlab
from bumpemu.util import ignore_exc
from bumpemu import debug


def _check_charger(port):
    with Powerlab(port) as pl:
        print('Reading options...')
        options = pl.read_options()
        print('Reading status...')
        status = pl.read_status()
        print('Reading presets (slow)...')
        presets = [preset for preset in pl.read_presets() if not preset.is_empty]
        print(options.greeting_line1.strip())
        print('Firmware: v%.2f' % (status.firmware_version / 100.0))
        print('Presets: %d' % len(presets))


def _print_preset(preset, preset_key):
    operation = preset_key[0].upper() + preset_key[1:]
    print('-' * 40)
    print(operation)
    print('-' * 40)
    print('  name: %s' % preset.name.strip())
    print('  preset_num: %s' % preset.preset_num)
    print('  chemistry: %s' % preset.chemistry)
    print('  charge_mamps: %s' % preset.charge_mamps)
    print('  max_charge_amps: %s' % preset.max_charge_amps)
    print('  charge_volts: %s' % preset.charge_volts)
    print('  discharge_mamps: %s' % preset.discharge_mamps)
    print('  discharge_volts: %s' % preset.discharge_volts)
    print('  num_parallel: %s' % preset.num_parallel)
    print('  num_cycles: %s' % preset.num_cycles)


def _print_presets(args, presets):
    keys = sorted(args.presets_config.keys())
    for key in keys:
        _print_preset(presets[args.presets_config[key]], key)


def _show_presets(args):
    with Powerlab(args.port) as pl:
        print('Reading presets (slow)...')
        presets = pl.read_presets()
        _print_presets(args, presets)


def _register_app_cb(logger, flag):
    logger.info('App registered')
    flag.registered = True


def _register_app_error_cb(logger, error, mainloop):
    logger.error('Failed to register app: %s', error)
    mainloop.quit()


def _register_ad_cb(logger, flag):
    logger.info('Advertisement registered')
    flag.registered = True


def _register_ad_error_cb(logger, error, mainloop):
    logger.error('Failed to register advertisement: %s', error)
    mainloop.quit()


def run(args, logger):
    try:
        if args.list_ports:
            for port in list_ports.comports():
                print(port)
            raise SystemExit(0)

        if args.check:
            _check_charger(args.port)
            raise SystemExit(0)

        if args.show_presets:
            _show_presets(args)
            raise SystemExit(0)

        presets = {constants.ChargerOperation[key.upper()]: val for key, val in args.presets_config.items()}

        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        bus = dbus.SystemBus()
        mainloop = GLib.MainLoop()

        # noinspection PyPep8Naming
        class registered_flag:
            registered = False

        app_flag = registered_flag()
        adv_flag = registered_flag()

        bluez_obj = bluez_dbus.get_bluez_obj(bus)
        app = None
        adv = None
        pl = None

        try:
            if not args.no_app_register:
                pl = Powerlab(args.port)
                service_manager = dbus.Interface(bluez_obj, bluez_dbus.GATT_MANAGER_IFACE)
                app = BumpEmulator(bus, 'bump_emulator', pl, args.battery, presets, args.status_interval)
                service_manager.RegisterApplication(
                    app.path,
                    {},
                    reply_handler=lambda: _register_app_cb(logger, app_flag),
                    error_handler=lambda ee: _register_app_error_cb(logger, ee, mainloop))
            if not args.no_advertise:
                adv_manager = dbus.Interface(bluez_obj, bluez_dbus.LE_ADVERTISING_MANAGER_IFACE)
                adv = UartAdvertisement(bus, 'bump_emulator', 0)
                adv_manager.RegisterAdvertisement(
                    adv.path,
                    {},
                    reply_handler=lambda: _register_ad_cb(logger, adv_flag),
                    error_handler=lambda ee: _register_ad_error_cb(logger, ee, mainloop))

            mainloop.run()
        finally:
            if adv_flag.registered:
                ignore_exc(func=lambda: adv_manager.UnregisterAdvertisement(adv.path))
                logger.info('Advertisement unregistered')
            if app_flag.registered:
                ignore_exc(func=lambda: service_manager.UnregisterApplication(app.path))
                logger.info('App unregistered')
            if pl:
                ignore_exc(func=pl.close)
    except KeyboardInterrupt:
        raise SystemExit(0)
    except SystemExit:
        pass
    except Exception as ex:
        logger.exception(ex)
        raise SystemExit(1)


def _positive_int(value):
    try:
        ivalue = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError('%s is an invalid positive int value' % value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError('%s is an invalid positive int value' % value)
    return ivalue


def _battery_file(value):
    if os.path.isfile(value):
        try:
            return Battery.from_yaml(value)
        except Exception as ex:
            reason = str(ex)
    else:
        reason = "not a file"
    raise argparse.ArgumentTypeError(reason)


def _presets_file(value):
    reason = ''
    if os.path.isfile(value):
        try:
            with open(value, 'r') as stream:
                data = load(stream, Loader=Loader)
                required = {'accurate', 'normal', 'fastest', 'storage', 'discharge', 'analyze', 'monitor'}
                missing = required.difference(data.keys())
                if missing:
                    raise argparse.ArgumentTypeError('the following are missing from the presets config '
                                                     'file: %s' % ', '.join(missing))
                ret = {}
                for key, val in data.items():
                    ret[key] = _positive_int(val) - 1
                return ret
        except argparse.ArgumentTypeError:
            raise
        except Exception as ex:
            reason = str(ex)
    else:
        reason = "not a file"
    raise argparse.ArgumentTypeError(reason)


def main():
    presets_yml = os.path.realpath('/etc/bumpemu/presets.yml')
    battery_yml = os.path.realpath('/etc/bumpemu/battery.yml')

    parser = argparse.ArgumentParser(usage='python3 -m bumpemu.main [options]',
                                     description='A bump controller for BLE that emulates a real bump controller.')
    parser.add_argument('--list-ports', action='store_true', help='List serial ports and exit.')
    parser.add_argument('-p', '--port', help='Set the serial port (default: auto search for port).')
    parser.add_argument('-c', '--check', action='store_true', help='Check the connection to the powerlab and exit.')
    parser.add_argument('-b', '--battery', type=_battery_file, metavar='YML', default=battery_yml,
                        help='Load a battery configuration from the given YAML file. (default: %s)' % battery_yml)
    parser.add_argument('--show-presets', action='store_true',
                        help='Read the presets from the powerlab, show the presets that will be used, and exit.')
    parser.add_argument('--presets-config', type=_presets_file, metavar='YML', default=presets_yml,
                        help='Set the presets configuration YAML file. (default: %s)' % presets_yml)
    parser.add_argument('-A', '--no-advertise', action='store_true',
                        help="Don't advertise on BLE.")
    parser.add_argument('-R', '--no-app-register', action='store_true',
                        help="Don't register the app with BLE.")
    parser.add_argument('--status-interval', type=_positive_int, default=1, metavar='SECONDS',
                        help=('Set the interval in seconds when status is retrieved from the charger. '
                              'Dev use only. (default: 1).'))
    parser.add_argument('-l', '--log-level', metavar='LEVEL', default='INFO',
                        help='Set the log level (default: INFO).')
    parser.add_argument('--log-serial', action='store_true', help='Turn on logging of the raw serial bytes.')
    parser.add_argument('--log-bluetooth', action='store_true', help='Turn on logging of the raw bluetooth bytes.')
    parser.add_argument('--log-status', action='store_true', help='Turn on logging of the charger status object.')
    pargs = parser.parse_args()

    loggr = logging.getLogger()
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s|%(levelname)s|%(filename)s:%(lineno)d|%(message)s')
    handler.setFormatter(formatter)
    loggr.addHandler(handler)
    loggr.setLevel(getattr(logging, pargs.log_level))

    debug.LOG_SERIAL = pargs.log_serial
    debug.LOG_BLUETOOTH = pargs.log_bluetooth
    debug.LOG_STATUS = pargs.log_status

    run(pargs, loggr)


if __name__ == '__main__':
    main()



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from bumpemu.controller import constants
from bumpemu.controller.serialize import append_uint16, append_int32, append_uint32


class ChargerStatus(object):
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self.port_number = 0
        self.schema_version = 6
        self.model_id = constants.ChargerModel.NONE
        self.comm_state = constants.CommState.COMM_DISCONNECTED
        self.mode_running = constants.ChargerMode.READY_TO_START
        self.error_code = 0
        self.chemistry = constants.Chemistry.NONE
        self._cell_count = 0
        self.estimated_fuel_level = 0
        self.estimated_minutes = 0
        self.amps = 0
        self.pack_volts = 0
        self.capacity_added = 0
        self.capacity_removed = 0
        self.cycle_timer = 0
        self.status_flags = 0
        self.rx_status_flags = 0
        self.operation_flags = 0
        self.power_reduced_reason = constants.ChargerPowerReducedReason.NONE
        self.supply_volts = 0
        self.supply_amps = 0
        self.cpu_temp = 0
        self.cell_volts = []
        self.cell_ir = []
        self.cell_bypass = []

    @property
    def cell_count(self):
        return self._cell_count

    @cell_count.setter
    def cell_count(self, val):
        self._cell_count = val
        self.cell_volts = [0] * val
        self.cell_ir = [0] * val
        self.cell_bypass = [0] * val

    def serialize(self):
        buf = bytearray()
        buf.append(self.port_number)
        buf.append(self.schema_version)
        # noinspection PyTypeChecker
        buf.append(self.model_id.value)
        # noinspection PyTypeChecker
        buf.append(self.comm_state.value)
        # noinspection PyTypeChecker
        buf.append(self.mode_running.value)
        buf.append(self.error_code)
        # noinspection PyTypeChecker
        buf.append(self.chemistry.value)
        buf.append(self.cell_count)
        buf.append(self.estimated_fuel_level)
        append_uint16(buf, self.estimated_minutes)
        append_int32(buf, self.amps)
        append_uint32(buf, self.pack_volts)
        append_uint32(buf, self.capacity_added)
        append_uint32(buf, self.capacity_removed)
        append_uint32(buf, self.cycle_timer)
        append_uint16(buf, self.status_flags)
        append_uint16(buf, self.rx_status_flags)
        buf.append(self.operation_flags)
        # noinspection PyTypeChecker
        buf.append(self.power_reduced_reason.value)
        append_uint32(buf, self.supply_volts)
        append_int32(buf, self.supply_amps)
        append_uint16(buf, self.cpu_temp)
        for ii in range(self.cell_count):
            append_uint16(buf, self.cell_volts[ii])
            append_uint16(buf, self.cell_ir[ii])
            buf.append(self.cell_bypass[ii])
        return buf



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bumpemu.util import byte_to_char, bits_from_int16, str_from_props, int16_to_bits, checksum


class Preset(object):
    CHEMISTRY = ['Empty', 'LiPo', 'Lith Ion', 'A123', 'LiMn', 'LiCo', 'NiCd', 'NiMH',
                 'Pb', 'LiFe (Chinese A123)', 'Primary (Dsch Only)', 'Supply (Low Voltage)',
                 'NiZn', 'LiHV']

    def __init__(self, data, preset_num):
        self._data = bytearray(data)
        self._preset_num = preset_num

    def raw_bytes(self):
        # noinspection PyAttributeOutsideInit
        self.checksum = self.calc_checksum()
        return bytes(self._data)

    @property
    def preset_num(self):
        return self._preset_num

    @property
    def is_require_balance_done_enabled(self):
        return bool(bits_from_int16(self._data, index=0, start_bit=0, end_bit=0))

    @is_require_balance_done_enabled.setter
    def is_require_balance_done_enabled(self, value):
        int16_to_bits(self._data, index=0, start_bit=0, end_bit=0, value=1 if value else 0)

    @property
    def is_require_all_charge_volts_enabled(self):
        return bool(bits_from_int16(self._data, index=0, start_bit=11, end_bit=11))

    @is_require_all_charge_volts_enabled.setter
    def is_require_all_charge_volts_enabled(self, value):
        int16_to_bits(self._data, index=0, start_bit=11, end_bit=11, value=1 if value else 0)

    @property
    def name(self):
        chars = []
        for ii in range(4, 32, 2):
            chars.append(byte_to_char(self._data[ii + 1]))
            chars.append(byte_to_char(self._data[ii]))
        return ''.join(chars)

    @name.setter
    def name(self, value):
        if len(value) > 28:
            raise Exception('name too long')
        if len(value) < 28:
            spaces = [' '] * (28 - len(value))
            value += ''.join(spaces)
        assert (len(value) == 28)
        for ii in range(4, 32, 2):
            self._data[ii + 1] = ord(value[ii - 4])
            self._data[ii] = ord(value[ii - 3])

    @property
    def charge_mamps(self):
        val = bits_from_int16(self._data, index=32, start_bit=4, end_bit=14)
        if val < 200:
            return val * 5
        else:
            return 1000 + ((val - 200) * 50)

    @charge_mamps.setter
    def charge_mamps(self, value):
        set_val = min(value, 40000)
        if set_val < 1000:
            set_val = int((set_val + 2.5) // 5) * 5
            set_val //= 5
        elif set_val >= 1000:
            set_val = ((set_val + 25) // 50) * 50
            set_val = ((set_val - 1000) // 50) + 200
        else:
            raise Exception('invalid value: %s' % value)
        int16_to_bits(self._data, index=32, start_bit=4, end_bit=14, value=set_val)

    @property
    def charge_volts(self):
        return bits_from_int16(self._data, index=34, start_bit=0, end_bit=9) / 200.0

    @charge_volts.setter
    def charge_volts(self, value):
        int16_to_bits(self._data, index=34, start_bit=0, end_bit=9, value=int(round(value * 200)))

    @property
    def discharge_mode(self):
        return bits_from_int16(self._data, index=84, start_bit=9, end_bit=11)

    @discharge_mode.setter
    def discharge_mode(self, val):
        int16_to_bits(self._data, index=84, start_bit=9, end_bit=11, value=val)

    @property
    def discharge_mamps(self):
        val = bits_from_int16(self._data, index=48, start_bit=0, end_bit=8)
        if val <= 100:
            return val * 10
        else:
            return 1000 + (val - 100) * 250

    @discharge_mamps.setter
    def discharge_mamps(self, value):
        if value <= 1000:
            value = ((value + 5) // 10) * 10
            value //= 10
        elif value > 1000:
            value = ((value + 125) // 250) * 250
            value = ((value - 1000) // 250) + 100
        else:
            raise Exception('invalid value: %s' % value)
        int16_to_bits(self._data, index=48, start_bit=0, end_bit=8, value=value)

    @property
    def discharge_volts(self):
        return bits_from_int16(self._data, index=98, start_bit=6, end_bit=14) / 100.0

    @discharge_volts.setter
    def discharge_volts(self, value):
        int16_to_bits(self._data, index=98, start_bit=6, end_bit=14, value=int(round(value * 100)))

    @property
    def is_store_charge_discharge(self):
        return bool(bits_from_int16(self._data, index=46, start_bit=12, end_bit=12))

    @is_store_charge_discharge.setter
    def is_store_charge_discharge(self, value):
        int16_to_bits(self._data, index=46, start_bit=12, end_bit=12, value=1 if value else 0)

    @property
    def is_end_cycling_with_discharge_enabled(self):
        return bool(bits_from_int16(self._data, index=46, start_bit=14, end_bit=14))

    @is_end_cycling_with_discharge_enabled.setter
    def is_end_cycling_with_discharge_enabled(self, value):
        int16_to_bits(self._data, index=46, start_bit=14, end_bit=14, value=1 if value else 0)

    @property
    def cool_down_time(self):
        return bits_from_int16(self._data, index=48, start_bit=10, end_bit=13)

    @cool_down_time.setter
    def cool_down_time(self, val):
        int16_to_bits(self._data, index=48, start_bit=10, end_bit=13, value=val)

    @property
    def cv_termination(self):
        return bits_from_int16(self._data, index=48, start_bit=14, end_bit=15)

    @cv_termination.setter
    def cv_termination(self, val):
        int16_to_bits(self._data, index=48, start_bit=14, end_bit=15, value=val)

    @property
    def is_balance_entire_charge_enabled(self):
        return bool(bits_from_int16(self._data, index=54, start_bit=15, end_bit=15))

    @is_balance_entire_charge_enabled.setter
    def is_balance_entire_charge_enabled(self, value):
        int16_to_bits(self._data, index=54, start_bit=15, end_bit=15, value=1 if value else 0)

    @property
    def beep_at_percent(self):
        return bits_from_int16(self._data, index=58, start_bit=11, end_bit=15) * 2 + 38

    @beep_at_percent.setter
    def beep_at_percent(self, val):
        int16_to_bits(self._data, index=58, start_bit=11, end_bit=15, value=(val - 38) // 2)

    @property
    def is_trickle_only(self):
        return bool(bits_from_int16(self._data, index=0, start_bit=5, end_bit=5))

    @is_trickle_only.setter
    def is_trickle_only(self, value):
        int16_to_bits(self._data, index=0, start_bit=5, end_bit=5, value=1 if value else 0)

    @property
    def is_balance_discharge_enabled(self):
        return bool(bits_from_int16(self._data, index=96, start_bit=8, end_bit=8))

    @is_balance_discharge_enabled.setter
    def is_balance_discharge_enabled(self, value):
        int16_to_bits(self._data, index=96, start_bit=8, end_bit=8, value=1 if value else 0)

    @property
    def chemistry_idx(self):
        return bits_from_int16(self._data, index=58, start_bit=6, end_bit=10)

    @chemistry_idx.setter
    def chemistry_idx(self, val):
        int16_to_bits(self._data, index=58, start_bit=6, end_bit=10, value=val)

    @property
    def chemistry(self):
        idx = self.chemistry_idx
        if idx >= len(self.CHEMISTRY):
            raise Exception('unknown battery type index: ' + idx)
        return self.CHEMISTRY[idx]

    # noinspection PyAttributeOutsideInit
    @chemistry.setter
    def chemistry(self, val):
        if isinstance(val, str):
            try:
                idx = self.CHEMISTRY.index(val)
            except ValueError:
                raise Exception('invalid chemistry: %s' % val)
        elif 0 <= val < len(self.CHEMISTRY):
            idx = val
        else:
            raise Exception('invalid chemistry: %s' % val)
        self.chemistry_idx = idx

    # 0: off
    # 1: constant current then constant voltage
    # 2: constant current
    @property
    def power_mode(self):
        return bits_from_int16(self._data, index=32, start_bit=0, end_bit=3)

    @power_mode.setter
    def power_mode(self, val):
        int16_to_bits(self._data, index=32, start_bit=0, end_bit=3, value=val)

    @property
    def is_requires_nodes_enabled(self):
        return bool(bits_from_int16(self._data, index=86, start_bit=13, end_bit=13))

    @is_requires_nodes_enabled.setter
    def is_requires_nodes_enabled(self, value):
        int16_to_bits(self._data, index=86, start_bit=13, end_bit=13, value=1 if value else 0)

    @property
    def auto_charge_rate(self):
        return bits_from_int16(self._data, index=0, start_bit=12, end_bit=15)

    @auto_charge_rate.setter
    def auto_charge_rate(self, val):
        int16_to_bits(self._data, index=0, start_bit=12, end_bit=15, value=val)

    @property
    def max_auto_charge_rate(self):
        return bits_from_int16(self._data, index=2, start_bit=10, end_bit=13)

    @max_auto_charge_rate.setter
    def max_auto_charge_rate(self, val):
        int16_to_bits(self._data, index=2, start_bit=10, end_bit=13, value=val)

    @property
    def is_use_fuel_enabled(self):
        return bool(bits_from_int16(self._data, index=0, start_bit=10, end_bit=10))

    @is_use_fuel_enabled.setter
    def is_use_fuel_enabled(self, value):
        int16_to_bits(self._data, index=0, start_bit=10, end_bit=10, value=1 if value else 0)

    @property
    def num_cycles(self):
        val = bits_from_int16(self._data, index=88, start_bit=10, end_bit=12)
        if val == 4:
            val = 5
        elif val == 5:
            val = 10
        elif val == 6:
            val = 20
        elif val == 7:
            val = 2 ** 32
        elif val > 7:
            raise Exception("unknown num cycles")
        return val

    @num_cycles.setter
    def num_cycles(self, val):
        if val not in (0, 1, 2, 3):
            if val == 5:
                val = 4
            elif val == 10:
                val = 5
            elif val == 20:
                val = 6
            elif val == 2 ** 32:
                val = 7
            else:
                raise Exception("unknown num cycles")
        int16_to_bits(self._data, index=88, start_bit=10, end_bit=12, value=val)

    @property
    def trickle_current_mamps(self):
        val = bits_from_int16(self._data, index=56, start_bit=9, end_bit=15)
        if val == 125:
            val = 1000
        elif val == 126:
            val = 2000
        elif val == 127:
            val = 3000
        else:
            val *= 5
        return val

    @trickle_current_mamps.setter
    def trickle_current_mamps(self, val):
        if val == 1000:
            val = 125
        elif val == 2000:
            val = 126
        elif val == 3000:
            val = 127
        elif val <= 620:
            val //= 5
        else:
            raise Exception('invalid value: %s' % val)
        int16_to_bits(self._data, index=56, start_bit=9, end_bit=15, value=val)

    @property
    def is_visible(self):
        return bool(bits_from_int16(self._data, index=32, start_bit=15, end_bit=15))

    @is_visible.setter
    def is_visible(self, value):
        int16_to_bits(self._data, index=32, start_bit=15, end_bit=15, value=1 if value else 0)

    @property
    def is_hide_empty_enabled(self):
        return bool(bits_from_int16(self._data, index=94, start_bit=15, end_bit=15))

    @is_hide_empty_enabled.setter
    def is_hide_empty_enabled(self, value):
        int16_to_bits(self._data, index=94, start_bit=15, end_bit=15, value=1 if value else 0)

    @property
    def is_locked(self):
        return bool(bits_from_int16(self._data, index=98, start_bit=15, end_bit=15))

    @is_locked.setter
    def is_locked(self, value):
        int16_to_bits(self._data, index=98, start_bit=15, end_bit=15, value=1 if value else 0)

    @property
    def num_parallel(self):
        return bits_from_int16(self._data, index=52, start_bit=8, end_bit=10) + 1

    @num_parallel.setter
    def num_parallel(self, value):
        if value <= 0:
            raise Exception('value must be >= 1')
        int16_to_bits(self._data, index=52, start_bit=8, end_bit=10, value=value - 1)

    @property
    def cv_timeout(self):
        return bits_from_int16(self._data, index=92, start_bit=5, end_bit=7)

    @cv_timeout.setter
    def cv_timeout(self, val):
        int16_to_bits(self._data, index=92, start_bit=5, end_bit=7, value=val)

    @property
    def charge_timeout(self):
        return bits_from_int16(self._data, index=52, start_bit=13, end_bit=15)

    @charge_timeout.setter
    def charge_timeout(self, val):
        int16_to_bits(self._data, index=52, start_bit=13, end_bit=15, value=val)

    @property
    def discharge_timeout(self):
        return bits_from_int16(self._data, index=54, start_bit=4, end_bit=6)

    @discharge_timeout.setter
    def discharge_timeout(self, val):
        int16_to_bits(self._data, index=54, start_bit=4, end_bit=6, value=val)

    @property
    def checksum(self):
        return (self._data[100] << 8) | self._data[101]

    @checksum.setter
    def checksum(self, value):
        self._data[100] = (value >> 8) & 0xff
        self._data[101] = value & 0xff

    @property
    def is_validated(self):
        return bool(bits_from_int16(self._data, index=36, start_bit=14, end_bit=15))

    @is_validated.setter
    def is_validated(self, value):
        int16_to_bits(self._data, index=36, start_bit=14, end_bit=15, value=1 if value else 0)

    @property
    def balance_mode(self):
        return bits_from_int16(self._data, index=82, start_bit=10, end_bit=13)

    @property
    def require_nodes(self):
        return bool(bits_from_int16(self._data, index=86, start_bit=13, end_bit=13))

    @property
    def fuel_curve(self):
        return [((self._data[ii] << 8) | self._data[ii + 1]) * 0.001111111 for ii in range(60, 82, 2)]

    @fuel_curve.setter
    def fuel_curve(self, val):
        ii = 60
        for vv in val:
            self._data[ii] = (vv >> 8) & 0xff
            self._data[ii + 1] = vv & 0xff
            ii += 2
            if ii == 82:
                break

    @property
    def max_charge_amps(self):
        val = bits_from_int16(self._data, index=34, start_bit=10, end_bit=15)
        if val == 0:
            return 0.25
        elif val == 1:
            return 0.5
        else:
            return val - 1

    @max_charge_amps.setter
    def max_charge_amps(self, val):
        if val < 1:
            val *= 100
            val = int(round(((val + 12.5) // 25) * 25))
            if val <= 25:
                val = 0
            elif val <= 50:
                val = 1
            else:
                val = 2
        else:
            val = min(int(round(val) + 1), 41)
        int16_to_bits(self._data, index=34, start_bit=10, end_bit=15, value=val)

    @property
    def is_empty(self):
        return (sum(self._data[:-2]) - self._data[94]) == 0

    def calc_checksum(self):
        return checksum(self._data[:-2], init=0x2d)

    def __str__(self):
        return str_from_props(self)



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import logging
from pprint import pformat

import dbus
import dbus.exceptions
import dbus.service

DBUS_OM_IFACE = 'org.freedesktop.DBus.ObjectManager'
BLUEZ_SERVICE_NAME = 'org.bluez'
LE_ADVERTISING_MANAGER_IFACE = 'org.bluez.LEAdvertisingManager1'
LE_ADVERTISEMENT_IFACE = 'org.bluez.LEAdvertisement1'
GATT_MANAGER_IFACE = 'org.bluez.GattManager1'
GATT_SERVICE_IFACE = 'org.bluez.GattService1'
GATT_CHRC_IFACE = 'org.bluez.GattCharacteristic1'
GATT_DESC_IFACE = 'org.bluez.GattDescriptor1'

PATH_BASE = '/com/example/service'


class InvalidArgsException(dbus.exceptions.DBusException):
    _dbus_error_name = 'org.freedesktop.DBus.Error.InvalidArgs'


class NotSupportedException(dbus.exceptions.DBusException):
    _dbus_error_name = 'org.bluez.Error.NotSupported'


class NotPermittedException(dbus.exceptions.DBusException):
    _dbus_error_name = 'org.bluez.Error.NotPermitted'


class InvalidValueLengthException(dbus.exceptions.DBusException):
    _dbus_error_name = 'org.bluez.Error.InvalidValueLength'


class FailedException(dbus.exceptions.DBusException):
    _dbus_error_name = 'org.bluez.Error.Failed'


def find_gatt_adapter(bus):
    remote_om = dbus.Interface(bus.get_object(BLUEZ_SERVICE_NAME, '/'), DBUS_OM_IFACE)
    objects = remote_om.GetManagedObjects()
    for o, props in objects.items():
        if GATT_MANAGER_IFACE in props.keys():
            return o
    return None


def find_adv_adapter(bus):
    remote_om = dbus.Interface(bus.get_object(BLUEZ_SERVICE_NAME, '/'), DBUS_OM_IFACE)
    objects = remote_om.GetManagedObjects()
    for o, props in objects.items():
        if LE_ADVERTISING_MANAGER_IFACE in props:
            return o
    return None


def get_bluez_obj(bus):
    adapter = find_gatt_adapter(bus)
    if not adapter:
        raise Exception('GattManager1 interface not found')
    return bus.get_object(BLUEZ_SERVICE_NAME, adapter)


class Advertisement(dbus.service.Object):
    def __init__(self, bus, path, index, advertising_type):
        self._logger = logging.getLogger(__name__)
        self._path = os.path.join('/org/bluez', path, 'advertisement') + str(index)
        self.bus = bus
        self.ad_type = advertising_type
        self.service_uuids = None
        self.manufacturer_data = None
        self.solicit_uuids = None
        self.service_data = None
        self.local_name = None
        self.include_tx_power = None
        self.data = None
        super(Advertisement, self).__init__(bus, self._path)

    @property
    def properties(self):
        properties = dict()
        properties['Type'] = self.ad_type
        if self.service_uuids is not None:
            properties['ServiceUUIDs'] = dbus.Array(self.service_uuids, signature='s')
        if self.solicit_uuids is not None:
            properties['SolicitUUIDs'] = dbus.Array(self.solicit_uuids, signature='s')
        if self.manufacturer_data is not None:
            properties['ManufacturerData'] = dbus.Dictionary(self.manufacturer_data, signature='qv')
        if self.service_data is not None:
            properties['ServiceData'] = dbus.Dictionary(self.service_data, signature='sv')
        if self.local_name is not None:
            properties['LocalName'] = dbus.String(self.local_name)
        if self.include_tx_power is not None:
            properties['IncludeTxPower'] = dbus.Boolean(self.include_tx_power)

        if self.data is not None:
            properties['Data'] = dbus.Dictionary(self.data, signature='yv')
        ret = {LE_ADVERTISEMENT_IFACE: properties}
        self._logger.debug('%s', pformat(ret))
        return ret

    @property
    def path(self):
        return dbus.ObjectPath(self._path)

    def add_service_uuid(self, uuid):
        if not self.service_uuids:
            self.service_uuids = []
        self.service_uuids.append(uuid)

    def add_solicit_uuid(self, uuid):
        if not self.solicit_uuids:
            self.solicit_uuids = []
        self.solicit_uuids.append(uuid)

    def add_manufacturer_data(self, manuf_code, data):
        if not self.manufacturer_data:
            self.manufacturer_data = dbus.Dictionary({}, signature='qv')
        self.manufacturer_data[manuf_code] = dbus.Array(data, signature='y')

    def add_service_data(self, uuid, data):
        if not self.service_data:
            self.service_data = dbus.Dictionary({}, signature='sv')
        self.service_data[uuid] = dbus.Array(data, signature='y')

    def add_local_name(self, name):
        if not self.local_name:
            self.local_name = ""
        self.local_name = dbus.String(name)

    def add_data(self, ad_type, data):
        if not self.data:
            self.data = dbus.Dictionary({}, signature='yv')
        self.data[ad_type] = dbus.Array(data, signature='y')

    # noinspection PyPep8Naming
    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, interface):
        if interface != LE_ADVERTISEMENT_IFACE:
            raise InvalidArgsException()
        return self.properties[LE_ADVERTISEMENT_IFACE]

    # noinspection PyPep8Naming
    @dbus.service.method(LE_ADVERTISEMENT_IFACE, in_signature='', out_signature='')
    def Release(self):
        pass


class Application(dbus.service.Object):
    """
    org.bluez.GattApplication1 interface implementation
    """

    def __init__(self, bus):
        self._logger = logging.getLogger(__name__)
        self._path = '/'
        self._services = []
        super(Application, self).__init__(bus, self._path)

    @property
    def path(self):
        return dbus.ObjectPath(self._path)

    def add_service(self, service):
        self._services.append(service)
        self._logger.debug('Added %s', service)

    # noinspection PyPep8Naming
    @dbus.service.method(DBUS_OM_IFACE, out_signature='a{oa{sa{sv}}}')
    def GetManagedObjects(self):
        response = {}

        for service in self._services:
            response[service.path] = service.properties
            for chrc in service.characteristics:
                response[chrc.path] = chrc.properties
                for desc in chrc.descriptors:
                    response[desc.path] = desc.properties

        return response

    def __str__(self):
        return 'Application<%s>: %s' % (self.uuid, self._path)


class Service(dbus.service.Object):
    """
    org.bluez.GattService1 interface implementation
    """

    def __init__(self, bus, path, index, uuid, primary):
        self._logger = logging.getLogger(__name__)
        self._path = os.path.join('/', path, 'service') + str(index)
        super(Service, self).__init__(bus, self._path)
        self.bus = bus
        self.uuid = uuid
        self.primary = primary
        self._characteristics = []

    @property
    def properties(self):
        return {
            GATT_SERVICE_IFACE: {
                'UUID': self.uuid,
                'Primary': self.primary,
                'Characteristics': dbus.Array(
                    self.characteristic_paths,
                    signature='o')
            }
        }

    @property
    def path(self):
        return dbus.ObjectPath(self._path)

    @property
    def characteristics(self):
        return self._characteristics

    @property
    def characteristic_paths(self):
        return [chrc.path for chrc in self._characteristics]

    def add_characteristic(self, characteristic):
        self._characteristics.append(characteristic)
        self._logger.debug('Added %s', characteristic)

    # noinspection PyPep8Naming
    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, interface):
        if interface != GATT_SERVICE_IFACE:
            raise InvalidArgsException()
        return self.properties[GATT_SERVICE_IFACE]

    def __str__(self):
        return 'Service<%s>: %s' % (self.uuid, self._path)


class Characteristic(dbus.service.Object):
    """
    org.bluez.GattCharacteristic1 interface implementation
    """

    def __init__(self, bus, index, uuid, flags, service):
        self._logger = logging.getLogger(__name__)
        self._path = service.path + '/char' + str(index)
        self.bus = bus
        self.uuid = uuid
        self.service = service
        self.flags = flags
        self._descriptors = []
        super(Characteristic, self).__init__(bus, self._path)

    @property
    def properties(self):
        return {
            GATT_CHRC_IFACE: {
                'Service': self.service.path,
                'UUID': self.uuid,
                'Flags': self.flags,
                'Descriptors': dbus.Array(
                    self.descriptor_paths,
                    signature='o')
            }
        }

    @property
    def path(self):
        return dbus.ObjectPath(self._path)

    @property
    def descriptors(self):
        return self._descriptors

    @property
    def descriptor_paths(self):
        return [desc.path for desc in self._descriptors]

    def add_descriptor(self, descriptor):
        self._descriptors.append(descriptor)
        self._logger.debug('Added %s', descriptor)

    # noinspection PyPep8Naming
    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, interface):
        if interface != GATT_CHRC_IFACE:
            raise InvalidArgsException()
        return self.properties[GATT_CHRC_IFACE]

    # noinspection PyPep8Naming
    @dbus.service.method(GATT_CHRC_IFACE, in_signature='a{sv}', out_signature='ay')
    def ReadValue(self, options):
        raise NotSupportedException()

    # noinspection PyPep8Naming
    @dbus.service.method(GATT_CHRC_IFACE, in_signature='aya{sv}')
    def WriteValue(self, value, options):
        raise NotSupportedException()

    # noinspection PyPep8Naming
    @dbus.service.method(GATT_CHRC_IFACE)
    def StartNotify(self):
        raise NotSupportedException()

    # noinspection PyPep8Naming
    @dbus.service.method(GATT_CHRC_IFACE)
    def StopNotify(self):
        raise NotSupportedException()

    # noinspection PyPep8Naming
    @dbus.service.signal(dbus.PROPERTIES_IFACE, signature='sa{sv}as')
    def PropertiesChanged(self, interface, changed, invalidated):
        pass

    def __str__(self):
        return 'Characteristic<%s>: %s' % (self.uuid, self._path)


class Descriptor(dbus.service.Object):
    """
    org.bluez.GattDescriptor1 interface implementation
    """

    def __init__(self, bus, index, uuid, flags, characteristic):
        self._logger = logging.getLogger(__name__)
        self._path = characteristic.path + '/desc' + str(index)
        self.bus = bus
        self.uuid = uuid
        self.flags = flags
        self.chrc = characteristic
        super(Descriptor).__init__(bus, self._path)

    @property
    def properties(self):
        return {
            GATT_DESC_IFACE: {
                'Characteristic': self.chrc.path,
                'UUID': self.uuid,
                'Flags': self.flags,
            }
        }

    @property
    def path(self):
        return dbus.ObjectPath(self._path)

    # noinspection PyPep8Naming
    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, interface):
        if interface != GATT_DESC_IFACE:
            raise InvalidArgsException()
        return self.properties[GATT_DESC_IFACE]

    # noinspection PyUnusedLocal
    # noinspection PyPep8Naming
    @dbus.service.method(GATT_DESC_IFACE, in_signature='a{sv}', out_signature='ay')
    def ReadValue(self, options):
        print('Default ReadValue called, returning error')
        raise NotSupportedException()

    # noinspection PyUnusedLocal
    # noinspection PyPep8Naming
    @dbus.service.method(GATT_DESC_IFACE, in_signature='aya{sv}')
    def WriteValue(self, value, options):
        print('Default WriteValue called, returning error')
        raise NotSupportedException()

    def __str__(self):
        return 'Descriptor<%s>: %s' % (self.uuid, self._path)



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from bumpemu.controller.state_machine.event import Event


class State(object):
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self._logger.debug('new state: %s', str(self))

    def on_event(self, event, emulator):
        pass

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.__class__.__name__


class DisconnectedState(State):
    def on_event(self, event, emulator):
        if event == Event.CONNECTED:
            return IdleState()
        return self


class IdleState(State):
    def on_event(self, event, emulator):
        if event == Event.HALT_FOR_SAFETY:
            return HaltForSafety(emulator)
        elif event == Event.STARTING:
            return StartingState()
        elif event == Event.CHARGING:
            return ChargingState()
        elif event == Event.DISCHARGING:
            return DischargingState()
        elif event == Event.MONITORING:
            return MonitoringState()
        elif event == Event.STOP:
            return StoppedState()
        elif event == Event.CHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.DISCHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.ERROR:
            return ErrorState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self


class StartingState(State):
    def on_event(self, event, emulator):
        if event == Event.HALT_FOR_SAFETY:
            return HaltForSafety(emulator)
        elif event == Event.CHARGING:
            return ChargingState()
        elif event == Event.DISCHARGING:
            return DischargingState()
        elif event == Event.IDLE:
            return IdleState()
        elif event == Event.STOP:
            return StoppedState()
        elif event == Event.CHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.DISCHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.ERROR:
            return ErrorState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self


class ChargingState(State):
    def on_event(self, event, emulator):
        if event == Event.STOP:
            return StoppedState()
        elif event == Event.CHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.DISCHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.ERROR:
            return ErrorState()
        elif event == Event.STARTING:
            return StartingState()
        elif event == Event.DISCHARGING:
            return DischargingState()
        elif event == Event.IDLE:
            return IdleState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self


class DischargingState(State):
    def on_event(self, event, emulator):
        if event == Event.STOP:
            return StoppedState()
        elif event == Event.CHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.DISCHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.ERROR:
            return ErrorState()
        elif event == Event.STARTING:
            return StartingState()
        elif event == Event.CHARGING:
            return ChargingState()
        elif event == Event.IDLE:
            return IdleState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self


class MonitoringState(State):
    def on_event(self, event, emulator):
        if event == Event.STARTING:
            return StartingState()
        elif event == Event.CHARGING:
            return ChargingState()
        elif event == Event.DISCHARGING:
            return DischargingState()
        elif event == Event.IDLE:
            return IdleState()
        elif event == Event.STOP:
            return StoppedState()
        elif event == Event.CHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.DISCHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.ERROR:
            return ErrorState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self


class CompletedState(State):
    def on_event(self, event, emulator):
        if event == Event.DISMISS:
            return IdleState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self


class StoppedState(State):
    def on_event(self, event, emulator):
        if event == Event.DISMISS:
            return IdleState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self


class ErrorState(State):
    def on_event(self, event, emulator):
        if event == Event.DISMISS:
            return IdleState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self


class HaltForSafety(State):
    def __init__(self, emulator):
        super(HaltForSafety, self).__init__()
        self._logger = logging.getLogger(__name__)
        try:
            emulator.clear_halt_for_safety()
        except Exception as ex:
            self._logger.exception(ex)

    def on_event(self, event, emulator):
        if event == Event.STOP:
            return StoppedState()
        elif event == Event.CHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.DISCHARGING_COMPLETE:
            return CompletedState()
        elif event == Event.ERROR:
            return ErrorState()
        elif event == Event.CHARGING:
            return ChargingState()
        elif event == Event.DISCHARGING:
            return DischargingState()
        elif event == Event.MONITORING:
            return MonitoringState()
        elif event == Event.IDLE:
            return IdleState()
        elif event == Event.DISCONNECTED:
            return DisconnectedState()
        return self



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os

from yaml import load, Loader
from bumpemu.controller import constants
from bumpemu.controller.serialize import append_uint16, append_array, append_str


class BatteryGroupNotify(object):
    def __init__(self, battery_group):
        self._logger = logging.getLogger(__name__)
        self.charger_port_number = 0
        self.battery_group = battery_group

    def serialize(self):
        buf = bytearray()
        buf.append(self.charger_port_number)
        self.battery_group.serialize(buf)
        return buf


class BatteryGroup(object):
    NFCID_COUNT = 8
    NFCID_LENGTH = 7

    def __init__(self, battery):
        self._logger = logging.getLogger(__name__)
        self.group_index = 0
        self.battery_count = battery.pack_count
        self.battery = battery
        self.index = 0
        self.nfc_ids = [[1, 2, 3, 4, 5, 6, 7]]
        for ii in range(1, self.NFCID_COUNT):
            self.nfc_ids.append([0] * self.NFCID_LENGTH)

    def serialize(self, buf=None):
        if buf is None:
            buf = bytearray()
        buf.append(self.group_index)
        buf.append(self.battery_count)
        self.battery.serialize(buf)
        for nfc_id in self.nfc_ids:
            append_array(buf, nfc_id)
        return buf


class Battery(object):
    REQUIRED = {'pref_operation', 'pref_charge_c_normal', 'pref_charge_c_fastest', 'pref_charge_c_accurate',
                'pref_discharge_c', 'cycle_count', 'internal_resistance', 'discharge_c_max', 'charge_c_max', 'capacity',
                'chemistry', 'cell_count', 'brand_name', 'max_cell_volts', 'min_cell_volts', 'pack_count',
                'storage_charge_volts', 'storage_discharge_volts',
                'pref_charge_c_discharge', 'pref_charge_c_storage', 'pref_charge_c_analyze', 'pref_charge_c_monitor'}

    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self.version = 2
        self.pref_operation = constants.ChargerOperation.NONE
        self.pref_charge_c_normal = 0
        self.pref_charge_c_fastest = 0
        self.pref_charge_c_accurate = 0
        self.pref_discharge_c = 0
        self.pref_fast_charge_delta = 0
        self.pref_discharge_delta = 0
        self.measured_fuel_table = [0] * 11
        self.measured_internal_resistance = 0
        self.measured_capacity = 0
        self.cycle_count = 0
        self.pref_accu_charge_delta = 0
        self.pref_norm_charge_delta = 0
        self.pref_store_charge_delta = 0
        self.pref_flags = 0
        self.battery_id = 0
        self.checksum = 0
        self.settings_version = 1
        self.internal_resistance = 0
        self.discharge_c_max = 0
        self.charge_c_max = 0
        self.capacity = 0
        self.chemistry = constants.ChargerOperation.NONE
        self.cell_count = 0
        self.brand_name = ''
        self.max_cell_volts = 0
        self.min_cell_volts = 0
        self.pack_count = 0

        # Not used by powerlab, only for emulator
        self.storage_charge_volts = 0
        self.storage_discharge_volts = 0
        self.pref_charge_c_discharge = 0
        self.pref_charge_c_storage = 0
        self.pref_charge_c_analyze = 0
        self.pref_charge_c_monitor = 0

    @classmethod
    def from_yaml(cls, yaml_file):
        battery = Battery()
        with open(yaml_file, 'r') as stream:
            data = load(stream, Loader=Loader)
            missing = cls.REQUIRED.difference(data.keys())
            if missing:
                raise Exception('the following are missing from the battery file: %s' % ', '.join(missing))
            if not data['brand_name']:
                raise Exception('"brand_name" minimum length is 1')
            if len(data['brand_name']) > 16:
                raise Exception('"brand_name" maximum length is 16')
            for var in vars(battery):
                if not var.startswith('_'):
                    val = data.get(var)
                    if val is not None:
                        setattr(battery, var, val)
        try:
            battery.pref_operation = constants.ChargerOperation[battery.pref_operation.upper()]
        except:
            raise Exception('"pref_operation" is invalid')
        try:
            battery.chemistry = constants.Chemistry[battery.chemistry.upper()]
        except:
            raise Exception('"chemistry" is invalid')
        return battery

    def serialize(self, buf=None):
        if buf is None:
            buf = bytearray()
        buf.append(self.version)
        # noinspection PyTypeChecker
        buf.append(self.pref_operation.value)
        append_uint16(buf, int(round(self.pref_charge_c_normal * 10)))
        append_uint16(buf, int(round(self.pref_charge_c_fastest * 10)))
        append_uint16(buf, int(round(self.pref_charge_c_accurate * 10)))
        append_uint16(buf, int(round(self.pref_discharge_c * 10)))
        buf.append(self.pref_fast_charge_delta)
        buf.append(self.pref_discharge_delta)
        append_array(buf, self.measured_fuel_table, append_func=append_uint16)
        append_uint16(buf, int(round(self.measured_internal_resistance * 100)))
        append_uint16(buf, self.measured_capacity)
        append_uint16(buf, self.cycle_count)
        buf.append(self.pref_accu_charge_delta)
        buf.append(self.pref_norm_charge_delta)
        buf.append(self.pref_store_charge_delta)
        buf.append(self.pref_flags)
        append_uint16(buf, self.battery_id)
        append_array(buf, [0] * 4)
        append_uint16(buf, self.checksum)
        buf.append(self.settings_version)
        append_uint16(buf, int(round(self.internal_resistance * 100)))
        append_uint16(buf, int(round(self.discharge_c_max)))
        append_uint16(buf, int(round(self.charge_c_max * 10)))
        append_uint16(buf, self.capacity)
        # noinspection PyTypeChecker
        buf.append(self.chemistry.value)
        buf.append(self.cell_count)
        append_str(buf, self.brand_name, 16)
        append_uint16(buf, int(round(self.max_cell_volts * 1000)))
        append_uint16(buf, int(round(self.min_cell_volts * 1000)))
        buf.append(self.pack_count)
        append_array(buf, [0] * 13)
        return buf

    def __str__(self):
        return os.linesep.join(
            ['%s: %s' % (var, getattr(self, var)) for var in sorted(vars(self)) if not var.startswith('_')])



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum
from bumpemu.controller import constants


class Event(Enum):
    NONE = 0
    CONNECTED = 1
    DISCONNECTED = 2
    IDLE = 3
    STARTING = 4
    CHARGING = 5
    DISCHARGING = 6
    MONITORING = 7
    STOP = 8
    CHARGING_COMPLETE = 9
    DISCHARGING_COMPLETE = 10
    DISMISS = 11
    ERROR = 12
    HALT_FOR_SAFETY = 13

    @classmethod
    def get_from_charge_status(cls, chg_status):
        mode = chg_status.mode
        if mode == constants.ChargerMode.READY_TO_START:
            return Event.IDLE
        elif mode == constants.ChargerMode.DETECTING_PACK.value:
            return Event.STARTING
        elif constants.ChargerMode.DETECTING_PACK.value < mode <= constants.ChargerMode.TRICKLE_CHARGING.value:
            if chg_status.is_charge_discharge_complete:
                return Event.CHARGING_COMPLETE
            else:
                return Event.CHARGING
        elif mode == constants.ChargerMode.DISCHARGING.value:
            if chg_status.is_charge_discharge_complete:
                return Event.DISCHARGING_COMPLETE
            else:
                return Event.DISCHARGING
        elif mode == constants.ChargerMode.MONITORING.value:
            return Event.MONITORING
        elif mode == constants.ChargerMode.HALT_FOR_SAFETY.value:
            return Event.HALT_FOR_SAFETY
        elif mode == constants.ChargerMode.ERROR.value:
            return Event.ERROR
        else:
            return Event.NONE



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import logging
import os
import sys

from bumpemu import debug
from bumpemu.charger.powerlab import Powerlab

import serial.tools.list_ports as list_ports


def run(args, logger):
    if args.list_ports:
        for port in list_ports.comports():
            print(port)
        return

    try:
        with Powerlab(args.port) as ser:
            if args.presets:
                all_presets = ser.read_presets()
                for pp in all_presets:
                    logger.debug('---------------------------')
                    logger.debug('Preset %d (cksum: %s): %s', pp.preset_num, pp.calc_checksum(), pp.name)
                    logger.debug('---------------------------')
                    logger.debug('%s%s', os.linesep, pp)
                if args.write:
                    ser.write_presets(all_presets)

            if args.status:
                ss = ser.read_status()
                logger.debug('%s%s', os.linesep, ss)
                if ss.error_code:
                    raise Exception('monitor command failed: %d' % ss.error_code)

            if args.options:
                opts = ser.read_options()
                logger.debug('%s%s', os.linesep, opts)
                if args.write:
                    ser.write_options(opts)

            if args.monitor:
                ser.command_monitor(num_parallel=args.num_parallel, use_bananas=not args.no_bananas)
            if args.enter:
                ser.command_enter()
            if args.charge:
                ser.command_charge(num_parallel=args.num_parallel, use_bananas=not args.no_bananas)
            if args.discharge:
                ser.command_discharge(num_parallel=args.num_parallel, use_bananas=not args.no_bananas)
            if args.cycle:
                ser.command_cycle(num_parallel=args.num_parallel, use_bananas=not args.no_bananas)
            if args.set_preset is not None:
                ser.command_set_active_preset(args.set_preset)
    except Exception as ex:
        logger.exception(ex)


def main():
    parser = argparse.ArgumentParser(description='')
    parser.add_argument('port', help='serial port')
    parser.add_argument('-l', '--log-level', default='INFO', help='set the log level (default: INFO)')
    parser.add_argument('-s', '--status', action='store_true', help='get status')
    parser.add_argument('-p', '--presets', action='store_true', help='get presets')
    parser.add_argument('-o', '--options', action='store_true', help='get options')
    parser.add_argument('-m', '--monitor', action='store_true', help='send the monitor command')
    parser.add_argument('-e', '--enter', action='store_true', help='send the enter command')
    parser.add_argument('-c', '--charge', action='store_true', help='send the charge command')
    parser.add_argument('-d', '--discharge', action='store_true', help='send the discharge command')
    parser.add_argument('-y', '--cycle', action='store_true', help='send the cycle command')
    parser.add_argument('-w', '--write', action='store_true', help='add write to the read')
    parser.add_argument('--set-preset', type=int, help='set the active preset')
    parser.add_argument('--num-parallel', type=int, default=1, help='set the number of parallel packs (default 1)')
    parser.add_argument('--no-bananas', action='store_true', help='turn off bananas')
    parser.add_argument('--list-ports', action='store_true', help='list serial ports')
    parser.add_argument('--log-serial', action='store_true', help='log the raw serial bytes')
    pargs = parser.parse_args()

    loggr = logging.getLogger()
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s|%(levelname)s|%(filename)s:%(lineno)d|%(message)s')
    handler.setFormatter(formatter)
    loggr.addHandler(handler)
    loggr.setLevel(getattr(logging, pargs.log_level))

    debug.LOG_SERIAL = pargs.log_serial

    run(pargs, loggr)


if __name__ == '__main__':
    main()



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct
import logging
# noinspection PyCompatibility
from queue import Queue
from threading import Thread

from bumpemu.circular_bytearray import CircularByteArray
from bumpemu.util import crc16
from bumpemu.controller import constants
from bumpemu.controller.messages.manual_start import ManualStart
from bumpemu import debug
from bumpemu.debug import print_bytes


class MessageHandler(object):
    def __init__(self, rx_chrc):
        self._logger = logging.getLogger(__name__)
        self._rx_chrc = rx_chrc
        self._buf = CircularByteArray(4096)
        self._message_handlers = {
            constants.MessageId.SELECTED_OPERATION_NOT.value: lambda xx: self._rx_chrc.selected_operation(
                port=xx[0], operation=xx[1]),
            constants.MessageId.OPERATION_START_CMD.value: lambda xx: self._rx_chrc.operation_start(port=xx[0]),
            constants.MessageId.OPERATION_STOP_CMD.value: lambda xx: self._rx_chrc.operation_stop(port=xx[0]),
            constants.MessageId.MONITOR_CMD.value: lambda xx: self._rx_chrc.monitor(port=xx[0]),
            constants.MessageId.OPERATION_CLEAR_ERROR_CMD.value: lambda xx: self._rx_chrc.clear_error(port=xx[0]),
            constants.MessageId.CONNECT_REQUEST.value: lambda xx: self._rx_chrc.connect_request(),
            constants.MessageId.CYCLE_GRAPH_GET.value: lambda xx: self._rx_chrc.cycle_graph_complete(),
            constants.MessageId.GET_DEVICE_INFO_CMD.value: lambda xx: self._rx_chrc.device_info(),
            constants.MessageId.DISMISS_CMD.value: lambda xx: self._rx_chrc.dismiss(port=xx[0], keep_setup=bool(xx[1])),
            constants.MessageId.MANUAL_OPERATION_CMD.value: lambda xx: self._rx_chrc.manual_operation(ManualStart(xx)),
            constants.MessageId.SET_BATTERY_GROUP_COUNT_CMD.value: lambda xx: self._rx_chrc.set_battery_group_count(
                port=xx[0], group_index=xx[1], count=xx[2]),
        }
        self._queue = Queue()
        self._thread = Thread(target=self._queue_processor, daemon=True)
        self._thread.start()

    def append(self, buf):
        self._queue.put(buf)

    def _queue_processor(self):
        while True:
            buf = self._queue.get()
            if debug.LOG_BLUETOOTH:
                print_bytes(self._logger, logging.DEBUG, buf, 'r')
            if self._buf.available() < len(buf):
                raise Exception('circular buffer is full')
            self._buf.append(buf)
            self._handle_messages()

    def _advance_to_next_preamble(self):
        while True:
            bb = self._buf.peek()
            if bb is None or bb == constants.Message.PREAMBLE_BYTE:
                break
            self._buf.advance(1)

    def _handle_messages(self):
        self._advance_to_next_preamble()

        buf_len = self._buf.size()
        while constants.Message.OVERHEAD <= buf_len:
            assert (self._buf.peek() == constants.Message.PREAMBLE_BYTE)
            payload_len_start_idx = self._buf.read_index + constants.Message.PAYLOAD_LEN_OFFSET
            payload_len = struct.unpack(constants.Message.PAYLOAD_LEN_FORMAT,
                                        self._buf[payload_len_start_idx:
                                                  payload_len_start_idx + constants.Message.PAYLOAD_LEN_BYTES])[0]
            message_size = payload_len + constants.Message.OVERHEAD
            if debug.LOG_BLUETOOTH:
                self._logger.debug('message_size: %d', message_size)
            if message_size > self._buf.capacity():
                self._logger.warning('message larger than circ buf size')
                self._advance_to_next_preamble()
                buf_len = self._buf.size()
                continue
            elif message_size <= buf_len:
                payload_start_idx = self._buf.read_index + constants.Message.PAYLOAD_OFFSET
                crc_start_idx = payload_start_idx + payload_len
                crc = struct.unpack(constants.Message.CRC_FORMAT,
                                    self._buf[crc_start_idx:crc_start_idx + constants.Message.CRC_BYTES])[0]
                calc_crc = crc16(self._buf[self._buf.read_index:crc_start_idx], init=constants.Message.CRC_SEED)
                if debug.LOG_BLUETOOTH:
                    self._logger.debug('crc: %s calc_crc: %s', hex(crc), hex(calc_crc))
                if crc == calc_crc:
                    message_id = self._buf[self._buf.read_index + constants.Message.MESSAGE_ID_OFFSET]
                    payload = self._buf[payload_start_idx:crc_start_idx]
                    self._buf.advance(constants.Message.OVERHEAD + payload_len)
                    self._handle_message(message_id, payload)
            else:
                # TODO: add timeout
                break
            self._advance_to_next_preamble()
            buf_len = self._buf.size()

    def _handle_message(self, message_id, payload):
        self._logger.debug('_handle_message - message_id: %s payload_len: %d', message_id, len(payload))
        handler = self._message_handlers.get(message_id)
        if handler:
            handler(payload)
        else:
            self._logger.debug('unhandled message id: %s', hex(message_id))



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import struct
from threading import RLock, Condition

import dbus
from gi.repository import GLib

from bumpemu.util import crc16
from bumpemu.controller import bluez_dbus
from bumpemu.controller.messages import bump_settings, charger_idle, charger_status, charger_settings, battery
from bumpemu.controller import constants
from bumpemu.controller.message_handler import MessageHandler
from bumpemu.controller.state_machine import state
from bumpemu.controller.state_machine.event import Event
from bumpemu.charger.powerlab import PowerlabException
from bumpemu import debug
from bumpemu.debug import print_bytes


class UartAdvertisement(bluez_dbus.Advertisement):
    def __init__(self, bus, path, index):
        super(UartAdvertisement, self).__init__(bus, path, index, 'peripheral')
        self.add_service_uuid(UartService.UUID)
        self.include_tx_power = True


class BumpEmulator(bluez_dbus.Application):
    def __init__(self, bus, path, charger, batt, presets, status_interval):
        super(BumpEmulator, self).__init__(bus)
        self._logger = logging.getLogger(__name__)
        self.add_service(UartService(bus, path, 0, charger, batt, presets, status_interval))


class UartService(bluez_dbus.Service):
    UUID = '6E400001-B5A3-F393-E0A9-E50E24DCCA9E'

    def __init__(self, bus, path, index, charger, batt, presets, status_interval):
        super(UartService, self).__init__(bus, path, index, self.UUID, True)
        self._logger = logging.getLogger(__name__)
        self._bus = bus
        self._rx_chrc = RxChrc(bus, 1, self, charger, batt, presets, status_interval)
        self.add_characteristic(TxChrc(bus, 0, self, self._rx_chrc))
        self.add_characteristic(self._rx_chrc)


class TxChrc(bluez_dbus.Characteristic):
    UUID = '6E400002-B5A3-F393-E0A9-E50E24DCCA9E'

    def __init__(self, bus, index, service, rx_chrc):
        super(TxChrc, self).__init__(bus, index, self.UUID, ['write-without-response'], service)
        self._logger = logging.getLogger(__name__)
        self._message_handler = MessageHandler(rx_chrc)

    def WriteValue(self, value, options):
        self._logger.debug('WriteValue')
        self._message_handler.append(bytes([bb for bb in value]))


def _modify_preset(preset, **kwargs):
    needs_update = False
    for key in kwargs.keys():
        if getattr(preset, key) != kwargs[key]:
            setattr(preset, key, kwargs[key])
            needs_update = True
    if preset.max_charge_amps != 40:
        preset.max_charge_amps = 40
        needs_update = True
    return needs_update


# noinspection PyAttributeOutsideInit
class RxChrc(bluez_dbus.Characteristic):
    UUID = '6E400003-B5A3-F393-E0A9-E50E24DCCA9E'
    MODEL_ID = 0x64
    FIRMWARE_VERSION = 408
    DEVICE_ID = [0, 1, 2, 3, 4, 5]
    DEVICE_NAME = 'BumpEmulator'

    def __init__(self, bus, index, service, charger, batt, presets, status_interval):
        super(RxChrc, self).__init__(bus, index, self.UUID, ['notify'], service)
        self._logger = logging.getLogger(__name__)
        self._notifying = False
        self._charger = charger
        self._battery = batt
        self._battery_group = battery.BatteryGroup(batt) if batt else None
        self._status_interval = status_interval * 1000
        self._lock = RLock()
        self._operation_to_preset_idx = presets
        self._selected_operation = batt.pref_operation if batt else None
        if self._selected_operation == constants.ChargerOperation.ANALYZE:
            raise Exception('analyze is not currently supported')
        self._running = False
        self._running_cond = Condition()
        self._bad_chemistry_error_code = 122  # unknown chemistry
        self._not_allowed_error_code = 49  # charge not allowed
        self._not_idle_error_code = 108  # preset loaded while charging
        self._op_not_set_error_code = 13  # preset is empty
        self._not_clearable_error_codes = {self._bad_chemistry_error_code}
        self._init()

    def _init(self):
        self._charger.close()
        self._state = state.DisconnectedState()
        self._charger_options = None
        self._presets = None
        self._active_preset = None
        self._forced_error_code = None
        self._disallow_operations = True
        self._no_status_count = 0

    def StartNotify(self):
        self._logger.debug('StartNotify')
        self._notifying = True

    def StopNotify(self):
        self._logger.debug('StopNotify')
        self._notifying = False
        self._logger.info('ble disconnected')

    def add_header(self, buf, message_id, payload):
        struct.pack_into(constants.Message.HEADER_FORMAT,
                         buf,
                         0,
                         constants.Message.PREAMBLE_BYTE,
                         self.MODEL_ID,
                         message_id,
                         len(payload))
        return constants.Message.HEADER_BYTES

    def _write(self, message_id, payload):
        if debug.LOG_BLUETOOTH:
            self._logger.debug('_write - notifying: %s', self._notifying)
        if self._notifying:
            buf = bytearray(len(payload) + 7)
            ii = self.add_header(buf, message_id, payload)
            for bb in payload:
                buf[ii] = bb
                ii += 1
            crc = crc16(buf[:-constants.Message.CRC_BYTES], init=constants.Message.CRC_SEED)
            struct.pack_into(constants.Message.CRC_FORMAT, buf, len(buf) - constants.Message.CRC_BYTES, crc)

            dbus_bytes = [dbus.Byte(bb) for bb in buf]
            if debug.LOG_BLUETOOTH:
                print_bytes(self._logger, logging.DEBUG, dbus_bytes, 'w')

            for ii in range(0, len(dbus_bytes), 40):
                self.PropertiesChanged(bluez_dbus.GATT_CHRC_IFACE, {'Value': dbus_bytes[ii:ii + 40]}, [])

    def connect_ack(self):
        self._logger.debug('connect_ack')
        buf = struct.pack('<HB', self.FIRMWARE_VERSION, 0)
        with self._lock:
            try:
                self._write(constants.MessageId.CONNECT_ACK.value, buf)
            except Exception as ex:
                self._logger.exception(ex)

    def connect_request(self):
        self._logger.debug('connect_request')

        # Wait for a status loop from a previous connection to exit (CCS app does not always send
        # StopNotify properly)
        assert self._notifying
        self._notifying = False
        try:
            with self._running_cond:
                while self._running:
                    if not self._running_cond.wait(self._status_interval * 2):
                        self._logger.error('timed out waiting for status loop to exit')
                        return
        finally:
            self._notifying = True

        self._init()

        # Don't allow any operations until we've done error checking and set everything up properly
        self._disallow_operations = True

        # Inform the app it is connected
        self.connect_ack()

        # Do one iteration of the status loop to set up initial state
        self.status_loop()

        # Start the status loop
        assert self._running
        GLib.timeout_add(self._status_interval, self.status_loop)

        self._logger.info('ble connected')

    def device_info(self):
        self._logger.debug('device_info')
        buf = bytearray()
        for bb in self.DEVICE_ID[:6]:
            buf.append(bb)
        for ch in self.DEVICE_NAME[:min(16, len(self.DEVICE_NAME))]:
            buf.append(ord(ch))
        for ii in range(len(buf), 22):
            buf.append(0x0)
        with self._lock:
            try:
                self._write(constants.MessageId.DEVICE_INFO.value, buf)
            except Exception as ex:
                self._logger.exception(ex)

    def select_charger(self):
        self._logger.debug('select_charger')
        buf = bytearray(1)
        buf[0] = 0
        with self._lock:
            try:
                self._write(constants.MessageId.SELECT_CHARGER_CMD.value, buf)
            except Exception as ex:
                self._logger.exception(ex)

    def bump_settings(self):
        self._logger.debug('bump_settings')
        settings = bump_settings.BumpSettings()
        settings.device_name = 'Bump Emulator'
        settings.presets_enabled = True
        name = '%s @%.1fA' % ('Battery' if self._charger_options.is_battery_enabled else 'DC Supply',
                              self._charger_options.supply_amps_limit)
        settings.set_power_source_params(index=0,
                                         name=name,
                                         typ=1 if self._charger_options.is_battery_enabled else 0,
                                         low_volts=self._charger_options.supply_cutoff_volts,
                                         max_amps=self._charger_options.supply_amps_limit)
        settings.set_power_source(port=0, index=0)
        settings.enable_charger_port(port=0)
        with self._lock:
            try:
                self._write(constants.MessageId.BUMP_SETTINGS.value, settings.serialize())
            except Exception as ex:
                self._logger.exception(ex)

    def battery_group(self):
        if self._battery_group:
            self._logger.debug('battery_group')
            bg = battery.BatteryGroupNotify(self._battery_group)
            try:
                self._write(constants.MessageId.BATTERY_GROUP_NOT.value, bg.serialize())
            except Exception as ex:
                self._logger.exception(ex)

    def charger_settings(self):
        if self._selected_operation:
            self._logger.debug('charger_settings')
            settings = charger_settings.ChargerSettings()
            settings.requested_operation = self._selected_operation
            settings.requested_chemistry = self._battery.chemistry
            settings.requested_cell_count = self._battery.cell_count
            settings.requested_ir = self._battery.internal_resistance
            settings.requested_capacity = self._battery.capacity * self._battery_group.battery_count
            operation_str = str(settings.requested_operation).split('.')[-1].lower()
            settings.requested_charge_c = getattr(self._battery, 'pref_charge_c_%s' % operation_str)
            settings.requested_discharge_c = 0
            if (settings.requested_operation in (constants.ChargerOperation.STORAGE,
                                                 constants.ChargerOperation.DISCHARGE,
                                                 constants.ChargerOperation.ANALYZE)):
                settings.requested_discharge_c = self._battery.pref_discharge_c
            settings.requested_charge_rate = settings.requested_charge_c * settings.requested_capacity
            settings.requested_discharge_rate = settings.requested_discharge_c * settings.requested_capacity
            settings.requested_charge_cutoff_cell_volts = self._battery.max_cell_volts
            settings.requested_discharge_cutoff_cell_volts = self._battery.min_cell_volts
            settings.requested_fuel_curve = self._battery.measured_fuel_table
            settings.multi_charger_mode = 0
            settings.power_supply_mode = constants.PowerSupplyMode(1 if self._charger_options.is_battery_enabled else 0)
            settings.use_balance_leads = True

            try:
                self._write(constants.MessageId.CHARGER_SETTINGS.value, settings.serialize())
            except Exception as ex:
                self._logger.exception(ex)

    @staticmethod
    def _mode_conversion(mode):
        if mode == 0:
            return constants.ChargerMode.READY_TO_START
        elif mode == 1:
            return constants.ChargerMode.DETECTING_PACK
        elif 2 <= mode <= 6:
            return constants.ChargerMode.CHARGING
        elif mode == 7:
            return constants.ChargerMode.TRICKLE_CHARGING
        elif mode == 8:
            return constants.ChargerMode.DISCHARGING
        elif mode == 9:
            return constants.ChargerMode.MONITORING
        elif mode == 10:
            return constants.ChargerMode.HALT_FOR_SAFETY
        elif mode == 11:
            return constants.ChargerMode.PACK_COOL_DOWN
        elif mode == 0x63:
            return constants.ChargerMode.ERROR
        raise Exception('unknown mode: %d' % mode)

    def _in_state(self, stat):
        return isinstance(self._state, stat)

    def _set_event(self, event):
        self._state = self._state.on_event(event, self)
        self.status_loop()

    def _set_forced_error(self, code):
        self._forced_error_code = code
        self.status_loop()

    def _check_preset_chemistries(self):
        if self._battery:
            for preset_num in self._operation_to_preset_idx.values():
                preset = self._presets[preset_num]
                if preset.chemistry_idx != self._battery.chemistry.value:
                    self._set_forced_error(self._bad_chemistry_error_code)
                    self._logger.error('preset %d "%s" is not the correct chemistry (%s != %s)',
                                       preset_num + 1, preset.name.strip(), self._battery.chemistry,
                                       preset.chemistry)

    def _update_presets(self):
        self._logger.debug('_update_presets')
        if self._battery and self._operation_to_preset_idx and self._in_state(state.IdleState):
            assert self._battery_group is not None

            presets_needing_update = []
            seen = set()
            for operation, preset_num in self._operation_to_preset_idx.items():
                operation_str = str(operation).split('.')[-1].lower()
                preset = self._presets[preset_num]
                if preset not in seen:
                    seen.add(preset)
                    charge_c = getattr(self._battery, 'pref_charge_c_%s' % operation_str)
                    new_vals = {
                        'auto_charge_rate': 0,
                        'charge_mamps': int(round(charge_c * self._battery.capacity)),
                        'discharge_mamps': int(round(self._battery.pref_discharge_c * self._battery.capacity)),
                        'num_parallel': self._battery_group.battery_count,
                    }
                    if operation_str == 'storage':
                        new_vals['charge_volts'] = self._battery.storage_charge_volts
                        new_vals['discharge_volts'] = self._battery.storage_discharge_volts
                    else:
                        new_vals['charge_volts'] = self._battery.max_cell_volts
                        new_vals['discharge_volts'] = self._battery.min_cell_volts
                        new_vals['num_cycles'] = self._battery.cycle_count

                    if _modify_preset(preset, **new_vals):
                        presets_needing_update.append(preset.preset_num + 1)

            if presets_needing_update:
                self._logger.debug('presets_needing_update: %s', presets_needing_update)
                if self.can_change_preset():
                    self._logger.info('writing presets')
                    self._charger.write_presets(self._presets, retries=2)
                    return False
            else:
                return False
        return True

    def check_preset(self, chg_status):
        if (self._active_preset is None or
                self._active_preset.preset_num != chg_status.active_preset):
            needs_update = True
            try:
                needs_update = self._update_presets()
            except Exception as ex:
                self._disallow_operations = True
                self._logger.exception(ex)

            if not needs_update and self._selected_operation is not None:
                assert self._operation_to_preset_idx is not None
                self._active_preset = self._presets[self._operation_to_preset_idx[self._selected_operation]]
                try:
                    self._charger.command_set_active_preset(self._active_preset.preset_num, retries=2)
                except Exception as ex:
                    self._disallow_operations = True
                    self._logger.exception(ex)
                else:
                    self._disallow_operations = self._battery is None or self._forced_error_code is not None
            else:
                self._disallow_operations = True

    def _charger_connected(self):
        assert self._charger_options

        self._logger.info('reading presets')
        try:
            self._presets = self._charger.read_presets(retries=2)
        except Exception as ex:
            self._logger.exception(ex)
        else:
            # Check that the preset chemistries match the specified battery chemistry
            self._check_preset_chemistries()

            # Inform the app of our parameters
            self.select_charger()
            self.bump_settings()
            self.charger_settings()
            self.battery_group()

            assert self._notifying

            self._state = self._state.on_event(Event.CONNECTED, self)
            self._logger.info('charger connected')

    def status_loop(self, force_idle=False):
        self._logger.debug('charger_status')
        chg_status = None
        with self._lock:
            if self._in_state(state.DisconnectedState):
                try:
                    self._charger_options = self._charger.connect()
                except PowerlabException:
                    pass
                else:
                    self._charger_connected()

            if not self._in_state(state.DisconnectedState):
                try:
                    chg_status = self._charger.read_status()
                except Exception as ex:
                    self._logger.exception(ex)
                    chg_status = None
                    self._no_status_count += 1
                    if self._no_status_count >= 5:
                        self._no_status_count = 0
                        self._charger_options = None
                        self._charger.close()
                        self._state = self._state.on_event(Event.DISCONNECTED, self)
                else:
                    self._no_status_count = 0

            if self._in_state(state.DisconnectedState):
                message_id = constants.MessageId.STATUS_IDLE_UPDATE_NOT2
                status = charger_idle.ChargerIdle()
                status.model_id = constants.ChargerModel.PL_8
                status.comm_state = constants.CommState.COMM_DISCONNECTED
                try:
                    self._write(message_id.value, status.serialize())
                except Exception as ex:
                    self._logger.exception(ex)

            elif chg_status:
                try:
                    if debug.LOG_STATUS:
                        self._logger.debug('%s', chg_status)
                    if self._forced_error_code is not None:
                        chg_status.error_code = self._forced_error_code
                        chg_status.mode = constants.ChargerMode.ERROR.value
                    event = Event.get_from_charge_status(chg_status)
                    self._state = self._state.on_event(event, self)

                    self.check_preset(chg_status)

                    is_idle_status = True
                    if isinstance(self._state, state.IdleState):
                        operation_flags = constants.ChargerOperationFlag.NONE
                    elif isinstance(self._state, state.HaltForSafety):
                        operation_flags = constants.ChargerOperationFlag.NONE
                        is_idle_status = False
                    elif isinstance(self._state, state.CompletedState):
                        operation_flags = constants.ChargerOperationFlag.COMPLETE
                        is_idle_status = False
                    elif isinstance(self._state, state.StoppedState):
                        operation_flags = constants.ChargerOperationFlag.STOPPED
                        is_idle_status = False
                    else:
                        operation_flags = constants.ChargerOperationFlag.NONE
                        is_idle_status = False

                    is_idle_status = (is_idle_status or force_idle) and self._forced_error_code is None

                    mode = self._mode_conversion(chg_status.mode)
                    self._logger.debug('state: %s is_idle_status: %s op_flags: %s mode: %s',
                                       self._state, is_idle_status, operation_flags, mode)

                    if is_idle_status:
                        message_id = constants.MessageId.STATUS_IDLE_UPDATE_NOT2
                        status = charger_idle.ChargerIdle()
                        status.firmware_version = chg_status.firmware_version
                    else:
                        message_id = constants.MessageId.STATUS_UPDATE_NOT2
                        status = charger_status.ChargerStatus()
                        status.mode_running = mode
                        status.error_code = chg_status.error_code
                        status.chemistry = constants.Chemistry(chg_status.chem8)
                        status.cell_count = chg_status.ch1_cells
                        status.estimated_fuel_level = int(round(chg_status.fuel_level / 10.0))
                        status.estimated_minutes = 0
                        status.amps = int(chg_status.avg_amps * 1000)
                        status.pack_volts = int(sum(chg_status.b_volts) * 1000)
                        status.capacity_added = int(round(chg_status.mah_in))
                        status.capacity_removed = int(round(chg_status.mah_out))
                        status.cycle_timer = chg_status.charge_seconds
                        status.status_flags = chg_status.status_flags
                        status.rx_status_flags = chg_status.rx_status_flags
                        if (isinstance(self._state, state.ChargingState) or
                                isinstance(self._state, state.DischargingState)):
                            if chg_status.lower_pwm_reason == 0 and chg_status.cv_started:
                                status.power_reduced_reason = constants.ChargerPowerReducedReason.OUTPUT_CV
                            else:
                                status.power_reduced_reason = constants.ChargerPowerReducedReason(
                                    chg_status.lower_pwm_reason)
                        else:
                            status.power_reduced_reason = constants.ChargerPowerReducedReason.NONE

                        if status.cell_count:
                            b_volts = chg_status.b_volts
                            mohm = chg_status.mohm
                            bp_pct = chg_status.bypass_percent
                            for ii in range(status.cell_count):
                                status.cell_volts[ii] = int(b_volts[ii] * 1000)
                                status.cell_ir[ii] = int(mohm[ii] * 100)
                                status.cell_bypass[ii] = int(round(bp_pct[ii]))

                    status.model_id = constants.ChargerModel.PL_8
                    status.comm_state = constants.CommState.COMM_CONNECTED
                    status.supply_volts = int(chg_status.supply_volts * 1000)
                    status.supply_amps = int(chg_status.supply_amps * 1000)
                    status.cpu_temp = int(round(chg_status.cpu_temp))
                    status.operation_flags = operation_flags.value

                    self._write(message_id.value, status.serialize())
                except Exception as ex:
                    self._logger.exception(ex)

            with self._running_cond:
                self._running = self._notifying
                if not self._running:
                    self._running_cond.notify()
            return self._notifying

    def cycle_graph_complete(self):
        self._logger.debug('cycle_graph_complete')
        buf = bytearray(1)
        buf[0] = 0
        with self._lock:
            try:
                self._write(constants.MessageId.CYCLE_GRAPH_GET_COMPLETE.value, buf)
            except Exception as ex:
                self._logger.exception(ex)

    def manual_operation(self, manual_start):
        self._logger.debug('manual_operation')
        self._logger.debug('%s', manual_start)
        with self._lock:
            self._logger.info('ignoring manual_operation: not supported')
            self._set_forced_error(self._not_allowed_error_code)

    def operation_start(self, port):
        self._logger.debug('operation_start(port=%d)', port)
        with self._lock:
            if self._disallow_operations:
                self._logger.info('ignoring operation_start: _disallow_operations==True')
                self._set_forced_error(self._not_allowed_error_code)
            elif not self._in_state(state.IdleState):
                self._logger.info('ignoring operation_start: not in idle state')
                self._set_forced_error(self._not_idle_error_code)
            elif not self._selected_operation:
                self._logger.info('ignoring operation_start: _selected_operation not set')
                self._set_forced_error(self._op_not_set_error_code)
            else:
                assert self._active_preset
                assert self._battery_group
                try:
                    if self._selected_operation == constants.ChargerOperation.DISCHARGE:
                        self._charger.command_discharge(self._battery_group.battery_count, retries=2)
                    else:
                        self._charger.command_charge(self._battery_group.battery_count, retries=2)
                except Exception as ex:
                    self._logger.exception(ex)

    def operation_stop(self, port):
        self._logger.debug('operation_stop(port=%d)', port)
        with self._lock:
            try:
                self._charger.command_enter(retries=2)
            except Exception as ex:
                self._logger.exception(ex)
            else:
                self._set_event(Event.STOP)

    def dismiss(self, port, keep_setup):
        self._logger.debug('dismiss(port=%d, keep_setup=%d)', port, keep_setup)
        with self._lock:
            try:
                self._charger.command_enter(retries=2)
            except Exception as ex:
                self._logger.exception(ex)
            else:
                self._set_event(Event.DISMISS)

    def clear_error(self, port):
        self._logger.debug('clear_error(port=%d)', port)
        with self._lock:
            try:
                if self._forced_error_code is None:
                    self._charger.command_enter(retries=2)
                if self._forced_error_code not in self._not_clearable_error_codes:
                    self._forced_error_code = None
            except Exception as ex:
                self._logger.exception(ex)
            else:
                self._set_event(Event.DISMISS)

    def set_battery_group_count(self, port, group_index, count):
        self._logger.debug('set_battery_group_count(port=%d, group_index=%d, count=%d)', port, group_index, count)
        with self._lock:
            if not self._in_state(state.IdleState):
                self._logger.info('ignoring set_battery_group_count: not in idle state')
                self._set_forced_error(self._not_idle_error_code)
            elif count != self._battery_group.battery_count:
                assert self._battery_group
                assert self._operation_to_preset_idx

                def update_preset_counts(cnt):
                    for preset_num in self._operation_to_preset_idx.values():
                        preset = self._presets[preset_num]
                        preset.num_parallel = cnt

                update_preset_counts(count)

                try:
                    self._logger.info('writing presets')
                    self._charger.write_presets(self._presets, retries=2)
                except Exception as ex:
                    update_preset_counts(self._battery_group.battery_count)
                    self._logger.exception(ex)
                else:
                    self._battery_group.battery_count = count
                    if self._active_preset:
                        assert self._battery_group.battery_count == self._active_preset.num_parallel

            self.battery_group()
            self.charger_settings()

    def monitor(self, port):
        self._logger.debug('monitor(port=%d)', port)
        with self._lock:
            if self._disallow_operations:
                self._logger.info('ignoring monitor: _disallow_operations==True')
                self._set_forced_error(self._not_allowed_error_code)
            elif not self._in_state(state.IdleState):
                self._logger.info('ignoring monitor: not in idle state')
                self._set_forced_error(self._not_idle_error_code)
            elif not self._active_preset:
                self._logger.info('ignoring monitor: _active_preset not set')
                self._set_forced_error(self._not_allowed_error_code)
            else:
                assert self._battery_group
                try:
                    self._charger.command_monitor(self._battery_group.battery_count, use_bananas=True, retries=2)
                except Exception as ex:
                    self._logger.exception(ex)

    def selected_operation(self, port, operation):
        self._logger.debug('selected_operation(port=%d, operation=%d)', port, operation)
        with self._lock:
            if self._disallow_operations:
                self._logger.info('ignoring selected_operation: _disallow_operations==True')
                self._set_forced_error(self._not_allowed_error_code)
            elif not self._in_state(state.IdleState):
                self._logger.info('ignoring selected_operation: not in idle state')
                self._set_forced_error(self._not_idle_error_code)
            else:
                new_op = constants.ChargerOperation(operation)
                if new_op == constants.ChargerOperation.ANALYZE:
                    self._logger.info('ignoring selected_operation: analyze not supported')
                    self._set_forced_error(self._not_allowed_error_code)
                else:
                    new_preset = self._presets[self._operation_to_preset_idx[new_op]]
                    try:
                        self._charger.command_set_active_preset(new_preset.preset_num, retries=2)
                    except Exception as ex:
                        self._logger.exception(ex)
                    else:
                        self._selected_operation = new_op
                        self._active_preset = new_preset
            self.charger_settings()

    def clear_halt_for_safety(self):
        self._logger.debug('clear_halt_for_safety')
        # lock should be held already
        self._charger.command_enter(retries=2)



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from time import sleep, time
from threading import Thread, Lock, Condition
import serial
import serial.tools.list_ports as list_ports

from bumpemu.circular_bytearray import CircularByteArray
from bumpemu import debug
from bumpemu.debug import print_bytes
from bumpemu.charger.status import Status
from bumpemu.charger.preset import Preset
from bumpemu.charger.options import Options
from bumpemu.util import swap_bytes, crc16, checksum, ignore_exc


class PowerlabException(Exception):
    pass


class ConnectFailedException(PowerlabException):
    pass


class VerifyException(PowerlabException):
    pass


class CrcException(VerifyException):
    pass


class ChecksumException(VerifyException):
    pass


def _command(cmd_str, nbytes=0):
    buf = bytearray(max(nbytes, len(cmd_str)))
    for ii in range(len(cmd_str)):
        buf[ii] = ord(cmd_str[ii])
    return buf


def _num_parallel_to_char(num):
    return chr(ord('l') + num - 1)


def _verify_preset_checksums(data):
    # every 510 bytes is checksummed
    for ii in range(0, 15):
        start = ii * 512
        end = start + 510
        calc_cksum = checksum(data[start:end], init=0xc8)
        check16 = (data[end] << 8) | data[end + 1]
        if calc_cksum != check16:
            raise ChecksumException('preset block checksum %d failed: cksum=%d calc_cksum=%d' %
                                    (ii, check16, calc_cksum))


def _prestart_offset(prestart_num):
    # every 510 bytes is checksummed
    return prestart_num * 102 + (prestart_num // 5) * 2


# We have to do this because the powerlab devs didn't implement flow control *sigh*
class SerialBuffer(object):
    READ_TIMEOUT = 1

    def __init__(self, ser):
        self._logger = logging.getLogger(__name__)
        self._ser = ser
        self._buf = CircularByteArray(48 * 1024)
        self._lock = Lock()
        self._cv = Condition(self._lock)
        self._stopped = False
        self._thread = Thread(target=self._read_loop, daemon=True)
        self._thread.start()

    def clear(self):
        self._buf.clear()

    def stop(self):
        self._stopped = True
        with self._cv:
            self._cv.notify()

    def read(self, nbytes, timeout=READ_TIMEOUT):
        with self._cv:
            while not self._stopped:
                start = time()
                data = self._buf.consume(nbytes)
                if data:
                    self._cv.notify()
                    return data
                timeout -= (time() - start)
                if not self._cv.wait(timeout=timeout):
                    return None  # timeout
            return None

    def _read_loop(self):
        self._ser.timeout = .1
        while not self._stopped:
            # noinspection PyBroadException
            try:
                data = self._ser.read(240)
            except:
                pass
            else:
                if data:
                    with self._cv:
                        while not self._stopped:
                            if self._buf.append(data):
                                self._cv.notify()
                                break
                            else:
                                self._cv.wait()


def _verify_cmd(cmd, buf):
    if len(buf) < len(cmd):
        raise VerifyException(cmd.decode('utf-8') + ' failed')
    if buf[:len(cmd)] != cmd:
        raise VerifyException(cmd.decode('utf-8') + ' failed')


def _verify_cmd_with_values(cmd, buf, byte_vals):
    _verify_cmd(cmd, buf)
    if buf[len(cmd):] != byte_vals:
        raise VerifyException(cmd.decode('utf-8') + ' failed')


def retry(func, num, interval=.1):
    while True:
        try:
            return func()
        except Exception as ex:
            if num <= 0:
                raise
            num -= 1
            if interval > 0:
                sleep(interval)


class Powerlab(object):
    READ_TIMEOUT = 1
    WRITE_TIMEOUT = 1

    def __init__(self, port):
        self._logger = logging.getLogger(__name__)
        self._port = port
        self._using_port = port
        self._ser = None
        self._serial_buffer = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, typ, value, traceback):
        self.close()

    @property
    def port(self):
        return self._using_port

    def connect(self, read_timeout=READ_TIMEOUT, write_timeout=WRITE_TIMEOUT):
        if not self._port:
            self._using_port = None
            for port in list_ports.comports():
                if port.description == 'FT232R USB UART':
                    self._using_port = port.device
                    break
            if not self._using_port:
                raise ConnectFailedException('no port found')

        self._logger.debug('connecting port:%s read_to:%f, write_to:%f', self._using_port, read_timeout, write_timeout)
        options = None
        baud_rates = (19200,)
        try:
            for baud_rate in baud_rates:
                ser = serial.PosixPollSerial(baudrate=baud_rate,
                                             timeout=read_timeout,
                                             write_timeout=write_timeout)
                ser.dtr = True
                ser.port = self._using_port
                ser.open()
                if ser.is_open:
                    self._ser = ser
                    if hasattr(self._ser, 'set_low_latency_mode'):
                        # noinspection PyUnresolvedReferences
                        self._ser.set_low_latency_mode(True)

                    retries = 3
                    while retries:
                        try:
                            options = self.read_options()
                            self._logger.info('connected to %s', self._using_port)
                            self._serial_buffer = SerialBuffer(self._ser)
                            return options
                        except Exception as ex:
                            retries -= 1
                            self._logger.debug(str(ex))

                    self._ser.close()
                    self._ser = None

        except Exception as ex:
            self._logger.debug(str(ex))
        raise ConnectFailedException()

    def close(self):
        if self._serial_buffer:
            self._serial_buffer.stop()
            self._serial_buffer = None
        if self._ser:
            if self._ser.is_open:
                ignore_exc(self._ser.reset_input_buffer)
                ignore_exc(self._ser.reset_output_buffer)
                ignore_exc(self._ser.close)
            self._ser = None
            self._logger.info('closed %s', self._using_port)

    def command_enter(self, num_parallel=1, retries=0):
        self._logger.debug('command_enter')
        return retry(lambda: self._send_cmd(num_parallel, 'E'), retries)

    def command_monitor(self, num_parallel, use_bananas=True, retries=0):
        self._logger.debug('command_monitor')
        return retry(lambda: self._send_cmd(num_parallel, 'M' if use_bananas else 'm'), retries)

    def command_charge(self, num_parallel, use_bananas=True, retries=0):
        self._logger.debug('command_charge')
        return retry(lambda: self._send_cmd(num_parallel, 'C' if use_bananas else 'c'), retries)

    def command_discharge(self, num_parallel, use_bananas=True, retries=0):
        self._logger.debug('command_discharge')
        return retry(lambda: self._send_cmd(num_parallel, 'D' if use_bananas else 'd'), retries)

    def command_cycle(self, num_parallel, use_bananas=True, retries=0):
        self._logger.debug('command_cycle')
        return retry(lambda: self._send_cmd(num_parallel, 'Y' if use_bananas else 'y'), retries)

    def command_set_active_preset(self, which, retries=0):
        self._logger.debug('command_set_active_preset %s', which)
        if which < 0 or which > 24:
            raise PowerlabException('invalid preset: %s' % which)
        def impl():
            write_cmd = _command('SelP' + chr(which))
            calc_crc = crc16([which], init=0x1114)
            self._write(write_cmd)
            resp = self._read(nbytes=len(write_cmd) + 2)
            crc = (resp[len(write_cmd)] << 8) | resp[len(write_cmd) + 1]
            if crc != calc_crc:
                raise CrcException('set preset %s failed: invalid CRC %s != %s' % (which, hex(crc), hex(calc_crc)))
        return retry(impl, retries)

    def read_status(self, retries=0):
        self._logger.debug('read_status')
        def impl():
            cmd = _command('Ram\0')
            self._write(cmd)
            resp = self._read(nbytes=153)
            self._verify_cmd_with_crc(cmd, resp, crc_index=151, crc_init=0x926)
            return Status(resp[len(cmd):151])
        return retry(impl, retries)

    def read_presets(self, retries=0):
        self._logger.debug('reading presets')
        def impl():
            cmd = _command('Prst')
            self._write(cmd)
            resp = self._read(nbytes=7686, timeout=7)
            _verify_cmd(cmd, resp)
            self._verify_crc(resp[4:], crc_index=7680, crc_init=0x18e4)
            _verify_preset_checksums(resp[4:])
            presets = []
            for preset_num in range(0, 75):
                offset = _prestart_offset(preset_num)
                presets.append(Preset(resp[4 + offset:4 + offset + 102], preset_num))
            return presets
        return retry(impl, retries)

    def write_presets(self, presets, retries=0):
        self._logger.debug('writing presets')
        def impl():
            write_cmd = _command('WrtP')
            ii = 1
            for preset in presets:
                preset.is_validated = not preset.is_empty
                write_cmd.extend(preset.raw_bytes())

                # every 510 bytes, compute and add checksum
                if (ii % 5) == 0:
                    block_num = ii // 5 - 1
                    start = 4 + block_num * 512
                    end = start + 510
                    assert (len(write_cmd) == end)
                    cksum = checksum(write_cmd[start:end], init=0xc8)
                    write_cmd.append(cksum >> 8)
                    write_cmd.append(cksum & 0xff)
                ii += 1

            if self._logger.isEnabledFor(logging.DEBUG):
                _verify_preset_checksums(write_cmd[4:])

            assert (len(write_cmd) == 7684)
            swap_bytes(write_cmd, start=4)
            calc_crc = crc16(write_cmd[4:], init=0x4d1)

            self._logger.debug('erase presets')
            cmd = _command('ErsP')
            self._write(cmd)
            resp = self._read(nbytes=6)
            _verify_cmd_with_values(cmd, resp, bytes([0x22, 0x1b]))

            sleep(.05)
            self._logger.debug('write presets')
            self._write(write_cmd, timeout=7)
            sleep(5.25)
            resp = self._read(nbytes=7686, timeout=7)
            if len(resp) != 7686:
                raise VerifyException('did not get expected response length: %d != %d' % (len(resp), 7686))
            crc = (resp[7684] << 8) | resp[7685]
            if crc != calc_crc:
                raise CrcException('write presets failed: invalid CRC %s != %s' % (hex(crc), hex(calc_crc)))
            self._logger.debug('presets write success')
        return retry(impl, retries)

    def read_options(self, retries=0):
        self._logger.debug('loading options')
        def impl():
            cmd = _command('PrsI')
            self._write(cmd)
            resp = self._read(nbytes=262)
            self._verify_cmd_with_crc(cmd, resp, crc_index=260, crc_init=0x342)
            return Options(resp[len(cmd):260])
        return retry(impl, retries)

    def write_options(self, options, retries=0):
        self._logger.debug('writing options')
        def impl():
            write_cmd = _command('WrtC')
            write_cmd.extend(options.raw_bytes()[128:192])

            assert (len(write_cmd) == 68)
            swap_bytes(write_cmd, start=4)
            calc_crc = crc16(write_cmd[4:], init=0xf5)

            self._logger.debug('erase options')
            cmd = _command('ErsC')
            self._write(cmd)
            resp = self._read(nbytes=6)
            _verify_cmd_with_values(cmd, resp, bytes([0x0d, 0x04]))

            self._logger.debug('write options')
            self._write(write_cmd)
            resp = self._read(nbytes=70)
            crc = (resp[68] << 8) | resp[69]
            if crc != calc_crc:
                raise CrcException('write options failed: invalid CRC %s != %s' % (hex(crc), hex(calc_crc)))
            self._logger.debug('options write success')
        return retry(impl, retries)

    def _read(self, nbytes=1, timeout=READ_TIMEOUT, retries=0):
        resp = None
        if self._serial_buffer:
            for ii in range(retries + 1):
                resp = self._serial_buffer.read(nbytes, timeout)
                if resp and len(resp) >= nbytes:
                    break
        else:
            self._ser.timeout = timeout
            for ii in range(retries + 1):
                resp = self._ser.read(nbytes)
                if resp and len(resp) >= nbytes:
                    break
        if retries and (resp is None or len(resp) < nbytes):
            raise VerifyException('_read did not get expected number of bytes')
        if debug.LOG_SERIAL:
            if resp:
                print_bytes(self._logger, logging.DEBUG, resp, 'r')
        return resp or []

    def _write(self, data, timeout=WRITE_TIMEOUT):
        self._ser.reset_output_buffer()
        self._ser.reset_input_buffer()
        if self._serial_buffer:
            self._serial_buffer.clear()
        self._ser.write_timeout = timeout
        self._ser.write(data)
        if debug.LOG_SERIAL:
            print_bytes(self._logger, logging.DEBUG, data, 'w')

    def _send_cmd(self, num_parallel, command_char):
        cmd = _command('Se' + _num_parallel_to_char(num_parallel) + command_char)
        self._write(cmd)
        resp = self._read(nbytes=6)
        _verify_cmd_with_values(cmd, resp, bytes([0x5, 0xdc]))

    def _verify_crc(self, buf, crc_index, crc_init):
        crc = (buf[crc_index] << 8) | buf[crc_index + 1]
        calc_crc = crc16(buf[:crc_index], crc_init)
        if debug.LOG_SERIAL:
            self._logger.debug('crc: %s calc_crc: %s', hex(crc), hex(calc_crc))
        if crc != calc_crc:
            raise CrcException('bad CRC')

    def _verify_cmd_with_crc(self, cmd, buf, crc_index, crc_init):
        _verify_cmd(cmd, buf)
        self._verify_crc(buf[len(cmd):], crc_index - len(cmd), crc_init)



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct


def append_array(buf, arr, append_func=None):
    if append_func:
        for ii in arr:
            append_func(buf, ii)
    else:
        for ii in arr:
            buf.append(ii)


def append_bool(buf, val):
    buf.append(1 if int(val) else 0)


def _append_struct(buf, val, fmt):
    for bb in struct.pack(fmt, int(val)):
        buf.append(bb)


def append_uint16(buf, val):
    _append_struct(buf, int(val), '<H')


def append_int32(buf, val):
    _append_struct(buf, int(val), '<l')


def append_uint32(buf, val):
    _append_struct(buf, int(val), '<L')


def append_str(buf, val, length):
    if val is None:
        val = ''
    for ii in val[:min(length, len(val))]:
        buf.append(ord(ii))
    for ii in range(length - len(val)):
        buf.append(0x0)


def read_uint16(buf):
    return struct.unpack('<H', buf)[0]



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

LOG_SERIAL = False
LOG_BLUETOOTH = False
LOG_STATUS = False


def print_bytes(logger, level, buf, prefix=''):
    if logger.isEnabledFor(level):
        ii = 0
        for bb in buf:
            if 0x20 <= bb <= 0x7e:
                logger.debug('%s [%d] %s: %s', prefix, ii, hex(bb), chr(bb))
            else:
                logger.debug('%s [%d] %s', prefix, ii, hex(bb))
            ii += 1



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import logging
from bumpemu.controller import constants
from bumpemu.controller.serialize import append_uint16, append_bool, read_uint16


class ManualStart(object):
    def __init__(self, buf=None):
        self._logger = logging.getLogger(__name__)
        if buf:
            self.deserialize(buf)
        else:
            self.charger_port_number = 0
            self.chemistry = constants.Chemistry.NONE
            self.cells = 0
            self.operation = constants.ChargerOperation.NONE
            self.cell_term_v = 0
            self.rate = 0
            self.balanced = True

    def serialize(self):
        buf = bytearray()
        buf.append(self.charger_port_number)
        # noinspection PyTypeChecker
        buf.append(self.chemistry.value)
        buf.append(self.cells)
        # noinspection PyTypeChecker
        buf.append(self.operation.value)
        append_uint16(buf, self.cell_term_v * 1000)
        append_uint16(buf, self.rate)
        append_bool(buf, self.balanced)
        return buf

    def deserialize(self, buf):
        self.charger_port_number = buf[0]
        self.chemistry = constants.Chemistry(buf[1])
        self.cells = buf[2]
        self.operation = constants.ChargerOperation(buf[3])
        self.cell_term_v = read_uint16(buf[4:6]) / 1000.0
        self.rate = read_uint16(buf[6:8])
        self.balanced = bool(buf[8])

    def __str__(self):
        return os.linesep.join(
            ['%s: %s' % (var, getattr(self, var)) for var in sorted(vars(self)) if not var.startswith('_')])



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct

from bumpemu.util import str_from_props


def _to_volts(val):
    return (val * 46.96) / 4095


_start_modes = {
    0: 'Charge Only',
    1: 'Discharge Only',
    2: 'Monitor',
    3: 'Cycle',
}


def _start_mode_to_str(mode):
    return _start_modes.get(mode, "Unknown: %d" % mode)


class Status(object):
    def __init__(self, data):
        self._data = data

    @property
    def firmware_version(self):
        return struct.unpack('>H', self._data[0:2])[0]

    @property
    def b_avg_adc(self):
        return struct.unpack('>HHHHHHHH', self._data[2:18])

    @property
    def b_volts(self):
        return [(val * 5.12) / 65536 for val in self.b_avg_adc]

    @property
    def charge_set(self):
        return struct.unpack('>H', self._data[20:22])[0]

    @property
    def l_supply_volts(self):
        return _to_volts(struct.unpack('>H', self._data[22:24])[0]) / 16

    @property
    def supply_volts(self):
        return _to_volts(struct.unpack('>H', self._data[24:26])[0])

    @property
    def cpu_temp(self):
        val = struct.unpack('>H', self._data[26:28])[0]
        return (((2.5 * val) / 4095.0) - 0.986) / 0.00355

    @property
    def avg_amps(self):
        return struct.unpack('>h', self._data[42:44])[0] / 600.0

    @property
    def status_flags(self):
        return struct.unpack('>H', self._data[44:46])[0]

    @property
    def safety_charge(self):
        return bool(self.status_flags & (1 << 0))

    @property
    def generate_fuel(self):
        return bool(self.status_flags & (1 << 5))

    @property
    def is_charge_discharge_complete(self):
        return bool(self.status_flags & (1 << 8))

    @property
    def is_reduce_amps(self):
        return bool(self.status_flags & (1 << 11))

    @property
    def show_vr(self):
        return bool(self.status_flags & (1 << 12))

    @property
    def node_current(self):
        return bool(self.status_flags & (1 << 14))

    @property
    def cold_weather(self):
        return bool(self.status_flags & (1 << 15))

    @property
    def rx_status_flags(self):
        return struct.unpack('>H', self._data[46:48])[0]

    @property
    def shunt_switch(self):
        return bool(self.rx_status_flags & (1 << 0))

    @property
    def dsch_enable(self):
        return bool(self.rx_status_flags & (1 << 1))

    @property
    def cd_pre_complete(self):
        return bool(self.rx_status_flags & (1 << 2))

    @property
    def regen_enable(self):
        return bool(self.rx_status_flags & (1 << 4))

    @property
    def fast_cell_avg(self):
        return bool(self.rx_status_flags & (1 << 5))

    @property
    def chg_enable(self):
        return bool(self.rx_status_flags & (1 << 6))

    @property
    def bp_enable(self):
        return bool(self.rx_status_flags & (1 << 7))

    @property
    def use_nodes(self):
        return bool(self.rx_status_flags & (1 << 8))

    @property
    def use_fuel(self):
        return bool(self.rx_status_flags & (1 << 9))

    @property
    def amps_low_range(self):
        return bool(self.rx_status_flags & (1 << 10))

    @property
    def amps_dsch_range(self):
        return bool(self.rx_status_flags & (1 << 11))

    @property
    def debug1(self):
        return struct.unpack('>h', self._data[48:50])[0]

    @property
    def _flags50(self):
        return struct.unpack('>H', self._data[50:52])[0]

    @property
    def high_temp(self):
        return bool(self._flags50 & (1 << 2))

    @property
    def cell_count_verified(self):
        return bool(self._flags50 & (1 << 12))

    @property
    def cell_vr(self):
        return [(((val * 5.12) / 4095) / 8) * 1000 for val in struct.unpack('>HHHHHHHH', self._data[52:68])]

    @property
    def vr_amps(self):
        return struct.unpack('>H', self._data[68:70])[0] / 600.0

    @property
    def vr_offset(self):
        return (((struct.unpack('>H', self._data[114:116])[0] * 5.12) / 4095) / 8) * 1000

    @property
    def ch1_cells(self):
        return self._data[132]

    @property
    def mohm(self):
        vals = [0] * 8
        if self.vr_amps > 0:
            vals[0] = (self.cell_vr[0] - self.vr_offset) / self.vr_amps
            for ii in range(1, len(vals)):
                if self.ch1_cells == ii:
                    vals[ii] = (self.cell_vr[ii] / self.vr_amps) - ((self.cell_vr[ii] / self.vr_amps) / 8.0)
                else:
                    vals[ii] = self.cell_vr[ii] / self.vr_amps
        return vals

    @property
    def _flags76(self):
        return struct.unpack('>H', self._data[76:78])[0]

    @property
    def checking_peak(self):
        return bool(self._flags76 & (1 << 0))

    @property
    def battery_24v_visible(self):
        return bool(self._flags76 & (1 << 3))

    @property
    def cv_started(self):
        return bool(self._flags76 & (1 << 4))

    @property
    def preset_good(self):
        return bool(self._flags76 & (1 << 5))

    @property
    def preset_flash_changed(self):
        return bool(self._flags76 & (1 << 6))

    @property
    def regen_possible(self):
        return bool(self._flags76 & (1 << 7))

    @property
    def regen_dsch_failed(self):
        return bool(self._flags76 & (1 << 8))

    @property
    def options_flash_changed(self):
        return bool(self._flags76 & (1 << 10))

    @property
    def supply_amps(self):
        return struct.unpack('>h', self._data[80:82])[0] / 150.0

    @property
    def batt_pos_avg_volts(self):
        return ((struct.unpack('>H', self._data[82:84])[0] * 46.96) / 4095) / 16

    @property
    def mode(self):
        return self._data[133]

    @mode.setter
    def mode(self, val):
        self._data[133] = val

    @property
    def mode_to_str(self):
        mode = self.mode
        if mode == 0:
            return 'idle'
        elif mode == 1:
            return 'detecting cells'
        elif mode == 2:
            return 'ch1 startup'
        elif mode == 3:
            return 'ch1/2 startup'
        elif mode == 6:
            if self.is_charge_discharge_complete:
                return 'charge complete'
            elif self.is_reduce_amps:
                return 'low voltage restore'
            else:
                return 'charging'
        elif mode == 7:
            if self.is_charge_discharge_complete:
                return 'charge complete'
            # elif chemistry == pb
            #    return 'finishing'
            else:
                return 'trickle charging'
        elif mode == 8:
            if self.is_charge_discharge_complete:
                return 'discharge complete'
            elif not self.regen_enable:
                return 'internal discharge'
            elif self.regen_enable:
                return 'regenerative discharge'
        elif mode == 9:
            return 'monitoring cells'
        elif mode == 10:
            return 'wait for button press'
        elif mode == 30:
            return 'slave mode'
        elif mode == 0x63:
            return 'safety code ' + self.error_code
        else:
            return 'unknown'

    @property
    def discharge_set(self):
        return struct.unpack('>H', self._data[92:94])[0]

    # if (this.LoadChemistryVal(Preset[SlaveNum]) == 11):
    #    this.MaxCellVolts[SlaveNum] = 0.001
    # else:

    @property
    def set_amps(self):
        if self.mode == 8:
            return self.discharge_set / 600.0
        else:
            return self.charge_set / 600.0

    @property
    def max_cell_volts(self):
        return ((struct.unpack('>H', self._data[74:76])[0] * 5.12) / 4095) / 16

    @property
    def avg_cell_volts(self):
        if self.use_nodes and self.ch1_cells:
            return sum(self.b_volts) / self.ch1_cells
        else:
            return self.max_cell_volts

    @property
    def avg_ir(self):
        if self.use_nodes and self.show_vr and self.ch1_cells:
            return sum(self.mohm) / self.ch1_cells
        else:
            return 0

    @property
    def slow_avg_amps(self):
        return struct.unpack('>H', self._data[116:118])[0] / 600

    # index = struct.unpack('>h', self._data[118:120])[0]
    # if index >= 0:
    #     ManualChgAmps[SlaveNum] = index;
    #     ManualAutoAmps[SlaveNum] = 0;
    # else:
    #     ManualChgAmps[SlaveNum] = 0
    #     ManualAutoAmps[SlaveNum] = (short)(0 - index);
    # SlavesFound[SlaveNum] = Module1.valbin2(inBytes, 120);
    # ManualDschAmps[SlaveNum] = Module1.valbin2S(inBytes, 0x7a);

    @property
    def bypass_pwm(self):
        return [bb for bb in self._data[124:132]]

    @property
    def bypass_percent(self):
        return [bp * 3.09375 for bp in self.bypass_pwm]

    @property
    def bypass_current(self):
        return [bp * 31.25 for bp in self.bypass_pwm]

    @property
    def error_code(self):
        return self._data[134]

    @error_code.setter
    def error_code(self, val):
        self._data[134] = val

    @property
    def chem8(self):
        return self._data[135]

    @property
    def packs(self):
        return self._data[136]

    @property
    def active_preset(self):
        num = self._data[137]
        if num > 74 or num < 0:
            return 0
        return num

    @property
    def screen_number(self):
        return self._data[139]

    @property
    def check_pack1_volts(self):
        return (struct.unpack('>b', self._data[140:141])[0] * 46.96) / 4095

    @property
    def fuel_offset(self):
        return int(round((self._data[141] * 5.12) / 4.095))

    @property
    def cycle_cnt(self):
        return self._data[142]

    @property
    def lower_pwm_reason(self):
        return self._data[143]

    # 0 = Charge only
    # 1 = Discharge only
    # 2 = Monitor
    # 3 = Cycle
    @property
    def start_mode(self):
        return self._data[144]

    @property
    def start_mode_str(self):
        return _start_mode_to_str(self.start_mode)

    @property
    def r_fail_reason(self):
        return self._data[145]

    @property
    def charge_seconds(self):
        secs = (self._data[28] << 8) | self._data[29]
        mins = (self._data[78] << 8) | self._data[79]
        if secs >= 0xfd1f:
            return (secs - 64800) + (mins * 60)
        else:
            return secs

    @property
    def mah_in(self):
        val = struct.unpack('>L', self._data[34:38])[0]
        if val > 0x7fffffff:
            val = 0
        if self.packs > 1:
            val /= float(self.packs)
        return val / 2160.0

    @property
    def mah_out(self):
        val = struct.unpack('>L', self._data[84:88])[0]
        if val > 0x7fffffff:
            val = 0
        if self.packs > 1:
            val /= float(self.packs)
        return val / 2160.0

    @property
    def fuel_level(self):
        val = struct.unpack('>h', self._data[38:40])[0]
        if val < 0:
            val = 0
        elif val > 1000:
            val = 1000
        return val

    @property
    def no_data_max(self):
        mode = self.mode
        if 6 <= mode <= 11:
            return 30
        else:
            return 3

    def __str__(self):
        return str_from_props(self)



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bumpemu.util import str_from_props, str_from_data, set_bit


# noinspection PyAttributeOutsideInit
class Options(object):
    def __init__(self, data):
        self._data = bytearray(data)

    @property
    def greeting_line1(self):
        return str_from_data(self._data, start=132, end=146)

    @property
    def greeting_line2(self):
        return str_from_data(self._data, start=148, end=158)

    @property
    def _flags128(self):
        return (self._data[128] << 8) | self._data[129]

    @_flags128.setter
    def _flags128(self, val):
        self._data[128] = val >> 8
        self._data[129] = val & 0xff

    @property
    def is_european_decimal(self):
        return bool(self._flags128 & (1 << 0))

    @property
    def is_button_click_enabled(self):
        return bool(self._flags128 & (1 << 1))

    @property
    def is_save_changes_enabled(self):
        return bool(self._flags128 & (1 << 2))

    @property
    def speaker_volume(self):
        return (self._flags128 & (0x7 << 4)) >> 4

    @property
    def cells_scroll_seconds(self):
        return (self._flags128 & (0x7 << 7)) >> 7

    @property
    def is_quick_start_enabled(self):
        return bool(self._flags128 & (1 << 10))

    @is_quick_start_enabled.setter
    def is_quick_start_enabled(self, val):
        self._flags128 = set_bit(self._flags128, 10, val)

    @property
    def regen_charge_voltage_in_to_pb(self):
        return (self._data[130] + 100) / 10.0

    @property
    def regen_amps_in_to_pb(self):
        return self._data[131] / 2.0

    @property
    def scroll_delay1(self):
        return self._data[164]

    @property
    def preset_name_scroll_speed(self):
        return self._data[165]

    @property
    def name_line2_secs(self):
        return self._data[166]

    @property
    def scroll_delay2(self):
        return self._data[167]

    @property
    def supply_cutoff_volts(self):
        return (self._data[168] + 100) / 10.0

    @property
    def supply_amps_limit(self):
        return self._data[169] / 2.0

    @property
    def _flags170(self):
        return (self._data[170] << 8) | self._data[171]

    @_flags170.setter
    def _flags170(self, val):
        self._data[170] = val >> 8
        self._data[171] = val & 0xff

    @property
    def is_cells_3_decimals_enabled(self):
        return bool(self._flags170 & (1 << 0))

    @property
    def is_quiet_charging(self):
        return bool(self._flags170 & (1 << 1))

    @is_quiet_charging.setter
    def is_quiet_charging(self, val):
        self._flags170 = set_bit(self._flags170, 1, val)

    @property
    def is_battery_enabled(self):
        return bool(self._flags170 & (1 << 4))

    @property
    def is_warn_50_dod_enabled(self):
        return bool(self._flags170 & (1 << 6))

    @property
    def is_regen_enabled(self):
        return bool(self._flags170 & (1 << 7))

    @property
    def is_choose_source_enabled(self):
        return bool(self._flags170 & (1 << 8))

    @is_choose_source_enabled.setter
    def is_choose_source_enabled(self, val):
        self._flags170 = set_bit(self._flags170, 8, val)

    @property
    def is_suppress_use_bananas_enabled(self):
        return bool(self._flags170 & (1 << 9))

    @is_suppress_use_bananas_enabled.setter
    def is_suppress_use_bananas_enabled(self, val):
        self._flags170 = set_bit(self._flags170, 9, val)

    @property
    def is_xh_node_wiring(self):
        return bool(self._flags170 & (1 << 10))

    @is_xh_node_wiring.setter
    def is_xh_node_wiring(self, val):
        self._flags170 = set_bit(self._flags170, 10, val)

    @property
    def is_network_disabled(self):
        return bool(self._flags170 & (1 << 11))

    @property
    def charge_done_beeps(self):
        return self._data[173]

    @property
    def battery_cutoff_volts(self):
        return (self._data[174] + 100) / 10.0

    @battery_cutoff_volts.setter
    def battery_cutoff_volts(self, val):
        self._data[174] = int(val * 10 - 100)

    @property
    def battery_amps_limit(self):
        return self._data[175] / 2.0

    @battery_amps_limit.setter
    def battery_amps_limit(self, val):
        self._data[175] = int(val * 2)

    @property
    def battery_type(self):
        return self._data[177]

    @property
    def checksum(self):
        return (self._data[190] << 8) | self._data[191]

    @checksum.setter
    def checksum(self, val):
        self._data[190] = (val >> 8) & 0xff
        self._data[191] = val & 0xff

    def raw_bytes(self):
        self.checksum = self.calc_checksum()
        return bytes(self._data)

    def calc_checksum(self):
        cksum = 0
        for ii in range(128, 186, 2):
            cksum += ((self._data[ii] << 8) | self._data[ii + 1])
        return cksum & 0xffff

    def __str__(self):
        return str_from_props(self)



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging


class CircularByteArray:
    def __init__(self, nbytes):
        self._logger = logging.getLogger(__name__)
        self._buf = bytearray(nbytes)
        self._write_idx = 0
        self._read_idx = -1

    @property
    def read_index(self):
        return self._read_idx + 1

    @property
    def write_index(self):
        return self._write_idx

    def capacity(self):
        return len(self._buf)

    def size(self):
        return self._write_idx - (self._read_idx + 1)

    def available(self):
        return len(self._buf) - self.size()

    def clear(self):
        self._read_idx = self._write_idx - 1

    def is_empty(self):
        return self.size() == 0

    def is_full(self):
        return self.available() == 0

    def append(self, data):
        if self.available() >= len(data):
            start = self._real_index(self._write_idx)

            jj = 0
            for ii in range(start, min(start + len(data), len(self._buf))):
                self._buf[ii] = data[jj]
                jj += 1
            for ii in range(0, len(data) - jj):
                self._buf[ii] = data[jj]
                jj += 1
            assert jj == len(data)

            self._write_idx += len(data)
            return True
        return False

    def consume(self, nbytes):
        if self.size() >= nbytes:
            data = self.__getitem__(slice(self.read_index, self.read_index + nbytes))
            self._read_idx += nbytes
            return data
        return None

    def advance(self, nbytes):
        self._read_idx += nbytes

    def peek(self):
        if self.size():
            return self.__getitem__(self.read_index)
        return None

    def __getitem__(self, val):
        if isinstance(val, slice):
            start = val.start
            if start is None:
                start = self.read_index
            if start < self.read_index or start >= self.write_index:
                raise IndexError

            stop = val.stop
            if stop <= self.read_index or stop > self.write_index:
                raise IndexError

            if stop <= start:
                return []

            nbytes = stop - start
            real_start = self._real_index(start)
            real_stop = self._real_index(stop)
            if real_start <= real_stop:
                return bytearray(self._buf[real_start:real_stop])
            else:
                buf = bytearray(nbytes)
                jj = 0
                for ii in range(real_start, len(self._buf)):
                    buf[jj] = self._buf[ii]
                    jj += 1
                for ii in range(0, real_stop):
                    buf[jj] = self._buf[ii]
                    jj += 1
                assert jj == len(buf)
                assert jj == nbytes
                return buf
        elif isinstance(val, int):
            if val < self.read_index or val >= self.write_index:
                raise IndexError
            return self._buf[self._real_index(val)]
        else:
            raise TypeError

    def _real_index(self, val):
        return val % len(self._buf)



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from bumpemu.controller import constants
from bumpemu.controller.serialize import append_uint16, append_uint32, append_int32


class ChargerIdle(object):
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self.port_number = 0
        self.model_id = constants.ChargerModel.NONE
        self.comm_state = constants.CommState.COMM_DISCONNECTED
        self.supply_volts = 0
        self.supply_amps = 0
        self.cpu_temp = 0
        self.operation_flags = 0
        self.firmware_version = 0

    def serialize(self):
        buf = bytearray()
        buf.append(self.port_number)
        # noinspection PyTypeChecker
        buf.append(self.model_id.value)
        # noinspection PyTypeChecker
        buf.append(self.comm_state.value)
        append_uint32(buf, self.supply_volts)
        append_int32(buf, self.supply_amps)
        append_uint16(buf, self.cpu_temp)
        buf.append(self.operation_flags)
        append_uint16(buf, self.firmware_version)
        return buf



#      Copyright (C) 2019  Frank Riley
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

REQUIRES = ['dbus-python>=1.2.8', 'PyGObject>=3.22.0', 'pyserial>=3.4', 'PyYAML>=3.13']

setuptools.setup(
    name="bumpemu",
    version="0.0.7",
    author="Frank Riley",
    author_email="fhriley@gmail.com",
    description="A bump controller emulator on Raspberry Pi Zero W.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fhriley/bumpemu",
    license='GNU GPLv3',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'bumpemu': ['config/presets.yml']},
    python_requires='>=3.5',
    install_requires=REQUIRES,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
    ],
    entry_points={
        'console_scripts': [
            'bumpemu-controller = bumpemu.main:main',
            'powerlab-tester = bumpemu.charger.tester:main'
        ]
    }
)
